# -*- coding: utf-8 -*-
"""Created on Thu Feb  2 10:37:05 2023.

@author: Barney

"""

from pathlib import Path
import tempfile

import pandas as pd
import pypandoc

from wsimod.core.core import WSIObj
from wsimod.nodes.nutrient_pool import NutrientPool

try:
    pypandoc.convert_text("# test string", "rst", format="md")
except OSError:
    pypandoc.download_pandoc(
        delete_installer=True, download_folder=tempfile.gettempdir()
    )


def get_classes():
    """

    Returns:

    """

    # Function to return all loaded subclasses of WSIObj (import arcs to include that)
    def all_subclasses(cls):
        """

        Args:
            cls:

        Returns:

        """
        return set(cls.__subclasses__()).union(
            [s for c in cls.__subclasses__() for s in all_subclasses(c)]
        )

    subclasses = all_subclasses(WSIObj)
    subclasses = {repr(x).split(".")[-1].replace("'>", ""): x for x in subclasses}
    return subclasses


def format_text(text):
    """

    Args:
        text:

    Returns:

    """
    # Convert docstring to a tidy format
    lines = text.strip().split("\n")
    lines = [line.strip().lstrip(" ") for line in lines]
    lines = [" - " + x.rstrip(" ") for x in " ".join(lines).split("- ") if len(x) > 0]
    return "<br>".join(lines)


# Identify which reference file to link modules to
reference_lookup = {}
for file in Path(__file__).parent.glob("reference-*"):
    with file.open() as f:
        for line in f:
            if ":::" in line:
                reference_lookup[line.replace("::: ", "").replace("\n", "")] = file.stem

# Iterate over classes
subclasses = get_classes()
subclasses["NutrientPool"] = NutrientPool
mytable = []
for subclass, obj in subclasses.items():
    # Get docstring
    otext = obj.__init__.__doc__
    if otext:
        # Only include docstrings with Key assumptions
        if "Key assumptions:" in otext:
            # Format reference
            rclass = ".".join(
                repr(obj).replace("<class '", "").replace("'>", "").split(".")
            )
            module = ".".join(rclass.split(".")[0:-1])
            subclass = "[`{0}`](./../{1}/#{2})".format(
                subclass, reference_lookup[module], rclass
            )

            # Format assumptions
            assum = format_text(
                otext.split("Key assumptions:")[1].split(
                    "Input data and parameter requirements:"
                )[0]
            )

            # Format inputs
            inputs = format_text(
                otext.split("Input data and parameter requirements:")[1]
            )

            # Append to table
            mytable.append(
                {"Component": subclass, "Assumptions": assum, "Data": inputs}
            )

# Convert to table
mytable = pd.DataFrame(mytable).sort_values(by="Component")
mytable["Assumptions"] = mytable["Assumptions"].str.replace("\n", "\n - ")
mytable["Component"] = mytable["Component"].str.replace("\n", "\n - ")
front_text = """
## Introduction\n
WSIMOD contains a variety of components to represent physical processes.
We recommend viewing the [API](#reference) for a detailed description of the
models included, however, we provide an overview of documented
components, their assumptions and required input data on this page.
\n\n

## Component Library
"""

# Write
with open(Path(__file__).parent / "component-library.md", "w") as f:
    f.write(front_text)
    f.write(mytable.to_markdown(index=False))



# -*- coding: utf-8 -*-
"""Created on Wed Apr  7 15:44:48 2021.

@author: Barney

Converted to totals on Thur Apr 21 2022
"""
from math import log10

from wsimod.core import constants


class WSIObj:
    """"""

    def __init__(self):
        """WSIObj is the base object of everything in WSIMOD. It is used to perform VQIP
        operations and mass balance checking behaviour.

        RSE has suggested that it would make more sense to leave VQIP operations as
        regular functions in a module or associated them with a VQIP class.

        Predefining empty_vqip_predefined in a class object is sensible though because
        it is the foundation of many operations, and copying a dict is many times
        quicker than copying a class.

        For now I will leave WSIObj as the base object, but this may change.
        """
        # Predefine empty concentrations because copying is quicker than defining
        self.empty_vqip_predefined = dict.fromkeys(constants.POLLUTANTS + ["volume"], 0)

    def empty_vqip(self):
        """Return a copy of the predefined empty vqip. All pollutants and volume
        initialised in a dict and set to 0.

        Returns:
            empty_vqip_predefined (dict): Copy of empty_vqip_predefined

        Examples:
            >>> obj = WSIObj()
            >>> obj.empty_vqip()
        """
        return self.empty_vqip_predefined.copy()

    def copy_vqip(self, t):
        """Wrapper to copy VQIP.

        Args:
            t (dict): A VQIP

        Returns:
            (dict): A copy of t
        """
        return t.copy()

    def blend_vqip(self, c1, c2):
        """Blends together two VQIPs that are assumed to have pollutant entries set as
        pollution concentrations, blending occurs with proportionate mixing.

        NOTE: VQIPs in WSIMOD in general store pollution as a total rather than a
        concentration. So you should only blend if you are doing it intentionally and
        know what you're doing. This won't do anything on VQIPs with 0 volume.

        Args:
            c1 (dict): A VQIP where pollutant entries are concentrations c2 (dict): A
            VQIP where pollutant entries are concentrations

        Returns:
            c (dict): A new VQIP where c1 and c2 have been proportionately blended
        """
        # Blend two vqips given as concentrations
        c = self.empty_vqip()

        c["volume"] = c1["volume"] + c2["volume"]
        if c["volume"] > 0:
            for pollutant in constants.POLLUTANTS:
                c[pollutant] = (
                    c1[pollutant] * c1["volume"] + c2[pollutant] * c2["volume"]
                ) / c["volume"]

        return c

    def sum_vqip(self, t1, t2):
        """Combines two VQIPs where pollutant entries are assumed to be given as mass.
        Volume and additive pollutants are summed while non additive pollutants are
        proportionately blended.

        Args:
            t1 (dict): A VQIP where pollutant entries are mass totals t2 (dict): A VQIP
            where pollutant entries are mass totals

        Returns:
            t (dict): A VQIP that is the sum of t1 and t2 (except for non-additive
                pollutants)

        Examples:
            >>> t1 = {'phosphate' : 0.25, 'volume' : 100, 'temperature' : 10}
            >>> t2 = {'phosphate' : 0.25, 'volume' : 10, 'temperature' : 15}
            >>> t = sum_vqip(t1, t2)
            >>> print(t)
            {'phosphate' : 0.5, 'volume' : 110, 'temperature' : 10.45}
        """
        # Sum two vqips given as totals
        t = self.copy_vqip(t1)
        t["volume"] += t2["volume"]
        for pollutant in constants.ADDITIVE_POLLUTANTS:
            t[pollutant] += t2[pollutant]

        if t["volume"] > 0:
            # Assume proportional blending of non additive pollutants
            for pollutant in constants.NON_ADDITIVE_POLLUTANTS:
                t[pollutant] = (
                    t2[pollutant] * t2["volume"] + t1[pollutant] * t1["volume"]
                ) / t["volume"]

        return t

    def concentration_to_total(self, c):
        """Convert a VQIP that has pollutant entries as concentrations into mass totals.

        Args:
            c (dict): A VQIP where pollutant entries are concentrations

        Returns:
            c (dict): A VQIP where pollutant entries are mass totals
        """
        c = self.copy_vqip(c)
        for pollutant in constants.ADDITIVE_POLLUTANTS:
            # Multiply concentration by volume to get mass for additive pollutants
            c[pollutant] *= c["volume"]
        return c

    def total_to_concentration(self, t):
        """Converts a VQIP that has pollutant entries as mass totals into
        concentrations. Note, that this won't work for VQIPs with 0 volume.

        Args:
            t (dict): A VQIP where pollutant entries are mass totals

        Returns:
            c (dict): A VQIP where pollutant entries are concentrations
        """
        c = self.copy_vqip(t)
        for pollutant in constants.ADDITIVE_POLLUTANTS:
            # Divide concentration by volume to get concentration for additive
            # pollutants
            c[pollutant] /= c["volume"]
        return c

    def extract_vqip(self, t1, t2):
        """Extract one VQIP from another where both VQIPs have pollutants as mass
        totals. Each volume and additive pollutant is directly subtracted.

        Args:
            t1 (dict): A VQIP where pollutant entries are mass totals to subtract
                from
            t2 (dict): A VQIP where pollutant entries are mass totals to subtract

        Returns:
            t (dict): A copy of t1 where each additive pollutant and volume has had
                t2 subtracted from it

        Examples:
            >>> t1 = {'phosphate' : 0.25, 'volume' : 100, 'temperature' : 10}
            >>> t2 = {'phosphate' : 0.25, 'volume' : 10, 'temperature' : 15}

            >>> t = extract_vqip(t1, t2)
            >>> print(t)
            {'phosphate' : 0, 'volume' : 90, 'temperature' : 10}
        """
        # TODO should probably be called 'subtract_vqip' TODO need to analyse uses of
        # this to see if it is sensible to do something for non additive
        t = self.copy_vqip(t1)
        # Directly subtract t2 from t1 for vol and additive pollutants
        for pol in constants.ADDITIVE_POLLUTANTS + ["volume"]:
            t[pol] -= t2[pol]

        return t

    def extract_vqip_c(self, c1, c2):
        """Extract one VQIP from another where both VQIPs have pollutants as
        concentrations. Operation performed for volume and additive pollutants.

        NOTE: VQIPs in WSIMOD in general store pollution as a total rather than a
        concentration. So you should only work with concentrations if you are doing it
        intentionally and know what you're doing.

        Args:
            c1 (dict): A VQIP where pollutant entries are concentrations to subtract
                from
            c2 (dict): A VQIP where pollutant entries are concentrations to subtract

        Returns:
            c (dict): A copy of c1 where each additive pollutant and volume has had
                c2 proportionately extracted from it
        """
        c = self.copy_vqip(c1)

        c1 = self.concentration_to_total(c1)
        c2 = self.concentration_to_total(c2)
        c["volume"] = c1["volume"] - c2["volume"]
        if c["volume"] > 0:
            for pollutant in constants.ADDITIVE_POLLUTANTS:
                # Subtract c2 from c1 for vol and additive pollutants
                c[pollutant] = (c1[pollutant] - c2[pollutant]) / c["volume"]

        return c

    def v_distill_vqip(self, t, v):
        """Directly remove a volume from a VQIP.

        Args:
            t (dict): A VQIP where pollutant entries are mass totals to remove
                volume from
            v (float): Volume to remove

        Returns:
            t (dict): Updated VQIP
        """
        # Distill v from t
        t = self.copy_vqip(t)
        t["volume"] -= v
        return t

    def v_distill_vqip_c(self, c, v):
        """Directly remove a volume from a VQIP, where pollutant entries are
        concentrations.

        NOTE: VQIPs in WSIMOD in general store pollution as a total rather than a
        concentration. So you should only work with concentrations if you are doing it
        intentionally and know what you're doing.

        Args:
            c (dict): A VQIP where pollutant entries are concentrations to remove
                volume from
            v (float): Volume to remove

        Returns:
            c (dict): Updated VQIP
        """
        # Distill v from c
        c = self.copy_vqip(c)
        d = self.empty_vqip()
        d["volume"] = -v
        c_ = self.blend_vqip(c, d)
        # Directly copy non additive pollutants
        for pollutant in constants.NON_ADDITIVE_POLLUTANTS:
            c_[pollutant] = c[pollutant]
        return c_

    def v_change_vqip(self, t, v):
        """Change the volume of a VQIP, where pollutants are mass totals, and update
        pollutant values in proportion to the change in volume.

        Args:
            t (dict): A VQIP where pollutant entries are mass totals to get
                pollutant concentrations from
            v (float): Volume from t to get proportionate pollutant values in

        Returns:
            (dict): A VQIP with v volume and pollutions in proportion to t

        Examples:
            You want to extract 10m3 from 100m3 of water (store), to do this you need to
            understand how much phosphate to extract in addition to volume.

            >>> store = {'volume' : 100, 'phosphate' : 0.25}
            >>> to_extract = v_change_vqip(store, 10)

            >>> print(to_extract)
            {'volume': 10, 'phosphate': 0.025}
        """
        t = self.copy_vqip(t)
        if t["volume"] > 0:
            # change all values of t by volume v in proportion to volume of t
            ratio = v / t["volume"]
            t["volume"] *= ratio
            for pol in constants.ADDITIVE_POLLUTANTS:
                t[pol] *= ratio

        else:
            # Assign volume directly
            t["volume"] = v
        return t

    def v_change_vqip_c(self, c, v):
        """Change the volume of a VQIP, where pollutants are concentrations.

        NOTE: VQIPs in WSIMOD in general store pollution as a total rather than a
        concentration. So you should only work with concentrations if you are doing it
        intentionally and know what you're doing.

        Args:
            c (dict): A VQIP where pollutant entries are concentrations v (float):
            Volume to change c's volume to

        Returns:
            c (dict): A new VQIP with volume udpated
        """
        # Change volume of vqip
        c = self.copy_vqip(c)
        c["volume"] = v
        return c

    def ds_vqip(self, t, t_):
        """Get difference between each additive pollutant and volume for VQIPs where
        pollutants are given as mass totals.

        Args:
            t (dict): A VQIP where pollutant entries are mass totals to subtract
                values from
            t_ (_type_): A VQIP where pollutant entries are mass totals to subtract

        Returns:
            ds (dict): Difference between t and t_ in mass totals

        Examples:
            >>> t1 = {'phosphate' : 0.25, 'volume' : 100, 'temperature' : 10}
            >>> t2 = {'phosphate' : 0.2, 'volume' : 90, 'temperature' : 9}

            >>> t = ds_vqip(t1, t2)
            >>> print(t)
            {'phosphate' : 0.05, 'volume' : 10, 'temperature' : 0}
        """
        ds = self.empty_vqip()
        for pol in constants.ADDITIVE_POLLUTANTS + ["volume"]:
            ds[pol] = t[pol] - t_[pol]
        return ds

    def ds_vqip_c(self, c, c_):
        """Get difference between each additive pollutant and volume for VQIPs where
        pollutants are given as concentrations but difference is given as mass totals.

        NOTE: VQIPs in WSIMOD in general store pollution as a total rather than a
        concentration. So you should only work with concentrations if you are doing it
        intentionally and know what you're doing.

        Args:
            c (dict): A VQIP where pollutant entries are concentrations to subtract
                values from
            c_ (_type_): A VQIP where pollutant entries are concentrations to
                subtract

        Returns:
            ds (dict): Difference between c and c_ in mass totals
        """
        ds = self.empty_vqip()
        ds["volume"] = c["volume"] - c_["volume"]
        for pol in constants.ADDITIVE_POLLUTANTS:
            ds[pol] = c["volume"] * c[pol] - c_["volume"] * c_[pol]
        # TODO what about non-additive ...
        return ds

    def compare_vqip(self, t1, t2):
        """Compare two VQIPs and check if the difference between each key is less ' than
        constants.FLOAT_ACCURACY.

        Args:
            t1 (dict): A VQIP t2 (dict): A VQIP

        Returns:
            bool: True if the difference is less for each key, False otherwise
        """
        reply = True
        for v in t1.keys():
            if abs(t1[v] - t2[v]) > constants.FLOAT_ACCURACY:
                reply = False
        return reply

    def mass_balance(self):
        """Call all mass balance functions and compare to see if discrepancy (i.e., if
        in_ != (out_ + ds_) for volume or for any additive pollutant).

        Comparison is performed in the magnitude of the largest value of in_, ds_ or
        out_. And so judgement should be exercised as to whether a mass balance has
        actually occurred

        Returns:
            in_ (dict): A VQIP of the total from mass_balance_in functions ds_ (dict): A
            VQIP of the total from mass_balance_ds functions out_ (dict): A VQIP of the
            total from mass_balance_out functions

        Raises:
            Message if mass balance does not close to constants.FLOAT_ACCURACY
        """
        # Iterate over mass_balance_in functions, summing values in in_
        in_ = self.empty_vqip()
        for f in self.mass_balance_in:
            in_ = self.sum_vqip(in_, f())

        # Iterate over mass_balance_out functions, summing values in out_
        out_ = self.empty_vqip()
        for f in self.mass_balance_out:
            out_ = self.sum_vqip(out_, f())

        # Iterate over mass_balance_ds functions, summing values in ds_
        ds_ = self.empty_vqip()
        for f in self.mass_balance_ds:
            ds_f = f()
            for v in constants.ADDITIVE_POLLUTANTS + ["volume"]:
                ds_[v] += ds_f[v]

        # Iterate over volume and additive pollutants to perform comparison
        for v in ["volume"] + constants.ADDITIVE_POLLUTANTS:
            # Find the largest value of in_, out_, ds_
            largest = max(in_[v], out_[v], ds_[v])

            if largest > constants.FLOAT_ACCURACY:
                # Convert perform comparison in a magnitude to match the largest value
                magnitude = 10 ** int(log10(largest))
                in_10 = in_[v] / magnitude
                out_10 = out_[v] / magnitude
                ds_10 = ds_[v] / magnitude
            else:
                in_10 = in_[v]
                ds_10 = ds_[v]
                out_10 = out_[v]

            if abs(in_10 - ds_10 - out_10) > constants.FLOAT_ACCURACY:
                # Print mass balance error Print actual difference rather than magnitude
                # comparison to enable user judgement

                print(
                    "mass balance error for {0} of {1} in {2}".format(
                        v, in_[v] - ds_[v] - out_[v], self.name
                    )
                )

        return in_, ds_, out_


class DecayObj(WSIObj):
    """"""

    # TODO - internet says this is a bad idea (diamond will occur when a Node - a type
    # of WSIObj inherits a DecayObj - also a type of WSIObj). The reason diamonds are
    # problems is because there can be conflicts in functions. But I don't want anyone
    # to overwrite WSIObj functions so I don't see an issue?
    def __init__(self, decays):
        """A WSIObj that has decay functions built in.

        Args:
            decays (dict): A dict of dicts containing a key for each pollutant that
            decays
                and, within that, a key for each parameter (a constant and exponent)

        Examples:
            The 'constant' parameter represents what proportion of an amount will
            decrease each time make_decay is called. Lower value will reduce decay.
            Bounded between 0 and 1. The 'exponent' parameter represents how temperature
            sensitive the decay is. The higher the value, the more pollution occurs at
            higher values. Values expected to vary between 1 (no temperature
            sensitivity) and 1.1 (high temperature sensitivity).

            >>> decays = {'phosphate' : {'constant' : 0.001, 'exponent' : 1.005}}

        Raises:
            Message if no access to temperature data
        """
        # Store decays
        self.decays = decays
        super().__init__()

        # Identify parent object to read temperature data
        if "parent" in dir(self):
            self.data_input_object = self.parent
        elif "in_port" in dir(self):
            self.data_input_object = self.in_port
        else:
            print("warning: decay object cannot access temperature data")

        self.total_decayed = self.empty_vqip()

    def make_decay(self, vqip):
        """Make decay, reading tempature and updating pollutant amounts. A wrapper for
        generic_temperature_decay.

        Args:
            vqip (dict): A VQIP to decay where pollutants are given as mass totals

        Returns:
            vqip_ (dict): A VQIP with pollutant amounts updated
        """
        # Read temperature data
        temperature = self.data_input_object.data_input_dict[
            ("temperature", self.data_input_object.t)
        ]
        # Make decay
        vqip_, diff = self.generic_temperature_decay(vqip, self.decays, temperature)
        # Update total_decayed for mass balance checking
        self.total_decayed = self.sum_vqip(self.total_decayed, diff)
        return vqip_

    def generic_temperature_decay(self, t, d, temperature):
        """Performs temperature sensitive pollutant decay calculations for a VQIP where
        pollutants are given as mass totals.

        Args:
            t (dict): A VQIP to decay where pollutants are given as mass totals d
            (dict): decays in a DecayObj temperature (float): temperature

        Returns:
            t (dict): A VQIP with updated pollutant values diff (dict): A VQIP storing
            the change in pollutant values (decreases
                stored as positive numbers)
        """
        t = self.copy_vqip(t)
        diff = self.empty_vqip()
        # Iterate over pollutants in d (keys)
        for pol, pars in d.items():
            # Perform calculation
            diff[pol] = t[pol] * min(
                pars["constant"]
                * pars["exponent"]
                ** (temperature - constants.DECAY_REFERENCE_TEMPERATURE),
                1,
            )
            # Update VQIP
            t[pol] -= diff[pol]

        return t, diff

    def generic_temperature_decay_c(self, c, d, temperature):
        """Performs temperature sensitive pollutant decay calculations for a VQIP where
        pollutants are given as concentrations.

        Args:
            c (dict): A VQIP to decay where pollutants are given as concentrations. d
            (dict): decays in a DecayObj temperature (float): temperature

        Returns:
            t (dict): A VQIP with updated pollutant values (pollutants as
                concentrations)
            diff (dict): A VQIP storing the change in pollutant values (decreases
                stored as positive numbers). Pollutants as mass totals.
        """
        c = self.copy_vqip(c)
        diff = self.empty_vqip()
        for pol, pars in d.items():
            diff[pol] = c[pol] * min(
                pars["constant"]
                * pars["exponent"]
                ** (temperature - constants.DECAY_REFERENCE_TEMPERATURE),
                1,
            )
            c[pol] -= diff[pol]

            diff[pol] *= c["volume"]
        return c, diff



# -*- coding: utf-8 -*-
"""Created on Tue Oct 18 10:35:51 2022.

@author: Barney

"""

# import pytest
import unittest
from math import exp
from unittest import TestCase

from wsimod.arcs.arcs import Arc
from wsimod.core import constants
from wsimod.nodes.land import (
    GardenSurface,
    GrowingSurface,
    ImperviousSurface,
    IrrigationSurface,
    Land,
    PerviousSurface,
    Surface,
)
from wsimod.nodes.nodes import Node
from wsimod.nodes.sewer import Sewer
from wsimod.nodes.storage import Reservoir
from wsimod.nodes.tanks import Tank, DecayTank
from wsimod.orchestration.model import to_datetime

from pathlib import Path


class MyTestClass(TestCase):
    def assertDictAlmostEqual(self, d1, d2, accuracy=19):
        """

        Args:
            d1:
            d2:
            accuracy:
        """
        for d in [d1, d2]:
            for key, item in d.items():
                d[key] = round(item, accuracy)
        self.assertDictEqual(d1, d2)

    def test_surface_read(self):
        constants.set_simple_pollutants()
        node = Node(name="", data_input_dict={("temperature", 1): 15})
        node.t = 1
        node.monthyear = 1

        surface = Surface(data_input_dict={("phosphate", 1): 0.1}, parent=node)

        self.assertEqual(surface.get_data_input("temperature"), 15)
        self.assertEqual(surface.get_data_input_surface("phosphate"), 0.1)

    def test_deposition(self):
        constants.set_simple_pollutants()
        d1 = {"volume": 0.1, "phosphate": 2, "temperature": 5}

        surface = Surface(parent=None, area=5, depth=0.1)
        surface.dry_deposition_to_tank(d1)
        self.assertDictAlmostEqual(d1, surface.storage)

        d2 = {"volume": 0.2, "phosphate": 4, "temperature": 5}
        surface.wet_deposition_to_tank(d1)
        self.assertDictAlmostEqual(d2, surface.storage)

    def test_atmospheric_dep(self):
        constants.set_simple_pollutants()
        node = Node(name="")
        node.monthyear = 1

        inputs = {("nhx-dry", 1): 0.1, ("noy-dry", 1): 0.4, ("srp-dry", 1): 0.2}

        surface = Surface(data_input_dict=inputs, parent=node, area=10)

        d1 = {"ammonia": 1, "nitrate": 4, "phosphate": 2, "volume": 0, "temperature": 0}
        d2 = {"volume": 0, "temperature": 0, "phosphate": 0}
        (r1, r2) = surface.atmospheric_deposition()
        self.assertDictAlmostEqual(d1, r1)
        self.assertDictAlmostEqual(d2, r2)

    def test_precip_dep(self):
        constants.set_simple_pollutants()
        node = Node(name="")
        node.monthyear = 1

        inputs = {("nhx-wet", 1): 0.1, ("noy-wet", 1): 0.4, ("srp-wet", 1): 0.2}

        surface = Surface(data_input_dict=inputs, parent=node, area=10)

        d1 = {"ammonia": 1, "nitrate": 4, "phosphate": 2, "volume": 0, "temperature": 0}
        d2 = {"volume": 0, "temperature": 0, "phosphate": 0}
        (r1, r2) = surface.precipitation_deposition()
        self.assertDictAlmostEqual(d1, r1)
        self.assertDictAlmostEqual(d2, r2)

    def test_run_surface(self):
        constants.set_default_pollutants()

        node = Land(name="", data_input_dict={("temperature", 1): 15})
        node.t = 1
        node.monthyear = 1

        inputs = {
            ("nhx-wet", 1): 0.1,
            ("noy-wet", 1): 0.4,
            ("srp-wet", 1): 0.2,
            ("nhx-dry", 1): 0.1,
            ("noy-dry", 1): 0.4,
            ("srp-dry", 1): 0.2,
        }

        decays = {
            "nitrite": {"constant": 0.005, "exponent": 1.005},
            "ammonia": {"constant": 0.05, "exponent": 1.05},
        }

        surface = Surface(data_input_dict=inputs, parent=node, area=10, decays=decays)

        surface.run()
        d1 = surface.empty_vqip()

        self.assertDictAlmostEqual(d1, node.running_outflow_mb)

        d1["phosphate"] = 4
        d1["ammonia"] = 2
        d1["nitrate"] = 8

        self.assertDictAlmostEqual(d1, node.running_inflow_mb)
        self.assertDictAlmostEqual(d1, surface.storage)

        surface.end_timestep()
        d1["ammonia"] = d1["ammonia"] * (1 - 0.05 * 1.05 ** (15 - 20))
        self.assertDictAlmostEqual(d1, surface.storage)

        surface.run()
        d1["phosphate"] = 8
        d1["ammonia"] = 2 + 2 * (1 - 0.05 * 1.05 ** (15 - 20))
        d1["nitrite"] = 2 * 0.05 * 1.05 ** (15 - 20)
        d1["nitrate"] = 16
        self.assertDictAlmostEqual(d1, surface.storage)

        surface.end_timestep()
        surface.run()
        d1["phosphate"] = 12
        d1["nitrate"] = 24 + d1["nitrite"] * (0.005 * 1.005 ** (15 - 20))
        d1["nitrite"] = d1["nitrite"] * (1 - 0.005 * 1.005 ** (15 - 20)) + d1[
            "ammonia"
        ] * (0.05 * 1.05 ** (15 - 20))
        d1["ammonia"] = (
            2
            + 2 * (1 - 0.05 * 1.05 ** (15 - 20))
            + 2 * (1 - 0.05 * 1.05 ** (15 - 20)) ** 2
        )

        self.assertDictAlmostEqual(d1, surface.storage, 16)

    def test_simple_dep(self):
        constants.set_simple_pollutants()
        d1 = {"phosphate": 2, "volume": 0, "temperature": 0}

        surface = Surface(pollutant_load=d1, area=1)

        (in_, out_) = surface.simple_deposition()

        d2 = {"volume": 0, "temperature": 0, "phosphate": 0}
        self.assertDictAlmostEqual(d1, in_)
        self.assertDictAlmostEqual(surface.storage, in_)
        self.assertDictAlmostEqual(d2, out_)

    def test_urban_precip(self):
        constants.set_simple_pollutants()
        inputs = {
            ("precipitation", 1): 0.1,
            ("et0", 1): 0.01,
            ("temperature", 1): 10,
            ("precipitation", 2): 0,
            ("et0", 2): 0.02,
            ("temperature", 2): 15,
        }

        node = Node(name="", data_input_dict=inputs)

        surface = ImperviousSurface(
            parent=node, area=1.5, et0_to_e=0.9, pore_depth=0.015
        )

        node.t = 1

        d1 = {"volume": 0.1 * 1.5, "temperature": 0, "phosphate": 0}
        d2 = {"phosphate": 0, "volume": 0.01 * 1.5 * 0.9, "temperature": 0}
        d3 = {"phosphate": 0, "temperature": 10, "volume": (0.1 - 0.01 * 0.9) * 1.5}
        (r1, r2) = surface.precipitation_evaporation()
        self.assertDictAlmostEqual(d1, r1, 17)
        self.assertDictAlmostEqual(d2, r2, 17)
        self.assertDictAlmostEqual(d3, surface.storage, 17)

        node.t = 2
        d3 = {"volume": 0, "temperature": 0, "phosphate": 0}
        d4 = {"phosphate": 0, "volume": 0.02 * 1.5 * 0.9, "temperature": 0}
        d5 = {
            "phosphate": 0,
            "temperature": 10,
            "volume": (0.1 - (0.01 + 0.02) * 0.9) * 1.5,
        }
        (r1, r2) = surface.precipitation_evaporation()
        self.assertDictAlmostEqual(d3, r1, 17)
        self.assertDictAlmostEqual(d4, r2, 17)
        self.assertDictAlmostEqual(d5, surface.storage, 17)

    def test_urban_push(self):
        constants.set_simple_pollutants()
        inputs = {
            ("precipitation", 1): 0.1,
            ("et0", 1): 0.01,
            ("temperature", 1): 10,
        }

        node = Node(name="", data_input_dict=inputs)
        node.t = 1
        sewer = Sewer(name="", capacity=2)
        _ = Arc(in_port=node, out_port=sewer, name="")
        surface = ImperviousSurface(
            parent=node, area=1.5, et0_to_e=0.9, pore_depth=0.015
        )
        _ = surface.precipitation_evaporation()
        _ = surface.push_to_sewers()

        d1 = {"volume": 0.015 * 1.5, "phosphate": 0, "temperature": 10}
        d2 = {
            "volume": (0.1 - 0.01 * 0.9 - 0.015) * 1.5,
            "phosphate": 0,
            "temperature": 10,
        }

        self.assertDictAlmostEqual(d1, surface.storage, 16)
        self.assertDictAlmostEqual(d2, sewer.sewer_tank.storage, 16)

    def test_perv_cmd(self):
        surface = PerviousSurface(
            parent="", depth=0.5, area=1.5, initial_storage=0.5 * 1.5 * 0.25
        )
        self.assertAlmostEqual((1 - 0.25) * 0.5, surface.get_cmd())

    def test_perv_cmd(self):
        surface = PerviousSurface(
            parent="", depth=0.5, area=1.5, initial_storage=0.5 * 1.5 * 0.25
        )
        self.assertAlmostEqual(0.25 * 0.5, surface.get_smc())

    def test_ihacres1(self):
        # Above field capacity
        constants.set_simple_pollutants()
        inputs = {
            ("precipitation", 1): 0.1,
            ("et0", 1): 0.01,
            ("temperature", 1): 10,
        }

        node = Node(name="", data_input_dict=inputs)

        surface = PerviousSurface(
            parent=node,
            depth=0.5,
            area=1.5,
            total_porosity=0.45,
            field_capacity=0.35,
            wilting_point=0.12,
            infiltration_capacity=0.4,
            surface_coefficient=0.04,
            percolation_coefficient=0.6,
            et0_coefficient=0.4,
            ihacres_p=12,
            initial_storage=0.4 * 0.5 * 1.5,
        )

        node.t = 1

        (r1, r2) = surface.ihacres()

        d1 = {"volume": 0.1 * 1.5, "phosphate": 0, "temperature": 0}

        d2 = {"volume": 0.01 * 1.5 * 0.4, "phosphate": 0, "temperature": 0}

        outflow_ = 0.1 * (
            1 - ((0.45 * 0.5 - 0.4 * 0.5) / (0.45 * 0.5 - 0.35 * 0.5)) ** 12
        )

        total_water_passing_through_soil = (0.1 - 0.01 * 0.4 - outflow_ * 0.04) * 1.5

        temperature = (0 * 0.4 * 0.5 * 1.5) + (
            total_water_passing_through_soil * 10
        ) / (total_water_passing_through_soil + 0.4 * 0.5 * 1.5)

        d3 = {
            "volume": outflow_ * (1 - 0.04) * 0.6 * 1.5,
            "phosphate": 0,
            "temperature": temperature,
        }

        d4 = {
            "volume": outflow_ * (1 - 0.04) * (1 - 0.6) * 1.5,
            "phosphate": 0,
            "temperature": temperature,
        }

        d5 = {
            "volume": (0.1 - outflow_ - 0.01 * 0.4 + 0.4 * 0.5) * 1.5,
            "phosphate": 0,
            "temperature": temperature,
        }

        d6 = {"volume": outflow_ * 0.04 * 1.5, "phosphate": 0, "temperature": 10}

        self.assertDictAlmostEqual(d1, r1)
        self.assertDictAlmostEqual(d2, r2)
        self.assertDictAlmostEqual(d3, surface.percolation, 15)
        self.assertDictAlmostEqual(d4, surface.subsurface_flow, 15)
        self.assertDictAlmostEqual(d5, surface.storage, 15)
        self.assertDictAlmostEqual(d6, surface.infiltration_excess)

    def test_ihacres2(self):
        # Below wilting point
        constants.set_simple_pollutants()
        inputs = {
            ("precipitation", 1): 0,
            ("et0", 1): 0.01,
            ("temperature", 1): 10,
        }

        node = Node(name="", data_input_dict=inputs)

        surface = PerviousSurface(
            parent=node,
            depth=0.5,
            area=1.5,
            total_porosity=0.45,
            field_capacity=0.35,
            wilting_point=0.12,
            infiltration_capacity=0.4,
            surface_coefficient=0.04,
            percolation_coefficient=0.6,
            et0_coefficient=0.4,
            ihacres_p=12,
            initial_storage=0.11 * 0.5 * 1.5,
        )

        node.t = 1

        (r1, r2) = surface.ihacres()

        evaporation_ = (
            0.01
            * 0.4
            * exp(2 * (1 - (0.45 * 0.5 - 0.11 * 0.5) / (0.45 * 0.5 - 0.12 * 0.5)))
            * 1.5
        )
        d1 = {"volume": 0, "phosphate": 0, "temperature": 0}

        d2 = {"volume": evaporation_, "phosphate": 0, "temperature": 0}

        d3 = {"volume": 0, "phosphate": 0, "temperature": 0}

        d4 = {
            "volume": 0.11 * 0.5 * 1.5 - evaporation_,
            "phosphate": 0,
            "temperature": 0,
        }

        d5 = {"volume": 0, "phosphate": 0, "temperature": 10}

        self.assertDictAlmostEqual(d1, r1)
        self.assertDictAlmostEqual(d2, r2)
        self.assertDictAlmostEqual(d3, surface.percolation, 15)
        self.assertDictAlmostEqual(d3, surface.subsurface_flow)
        self.assertDictAlmostEqual(d4, surface.storage, 15)
        self.assertDictAlmostEqual(d5, surface.infiltration_excess)

    def test_ihacres3(self):
        # Infiltration excess
        constants.set_simple_pollutants()
        inputs = {
            ("precipitation", 1): 0.1,
            ("et0", 1): 0.01,
            ("temperature", 1): 10,
        }

        node = Node(name="", data_input_dict=inputs)

        surface = PerviousSurface(
            parent=node,
            depth=0.5,
            area=1.5,
            total_porosity=0.45,
            field_capacity=0.35,
            wilting_point=0.12,
            infiltration_capacity=0.05,
            surface_coefficient=0.04,
            percolation_coefficient=0.6,
            et0_coefficient=0.4,
            ihacres_p=12,
            initial_storage=0.4 * 0.5 * 1.5,
        )

        node.t = 1

        (r1, r2) = surface.ihacres()

        d1 = {"volume": 0.1 * 1.5, "phosphate": 0, "temperature": 0}

        d2 = {"volume": 0.01 * 1.5 * 0.4, "phosphate": 0, "temperature": 0}

        outflow_ = 0.05 * (
            1 - ((0.45 * 0.5 - 0.4 * 0.5) / (0.45 * 0.5 - 0.35 * 0.5)) ** 12
        )

        total_water_passing_through_soil = (0.05 - 0.01 * 0.4 - outflow_ * 0.04) * 1.5

        temperature = (0 * 0.4 * 0.5 * 1.5) + (
            total_water_passing_through_soil * 10
        ) / (total_water_passing_through_soil + 0.4 * 0.5 * 1.5)

        d3 = {
            "volume": outflow_ * (1 - 0.04) * 0.6 * 1.5,
            "phosphate": 0,
            "temperature": temperature,
        }

        d4 = {
            "volume": outflow_ * (1 - 0.04) * (1 - 0.6) * 1.5,
            "phosphate": 0,
            "temperature": temperature,
        }

        d5 = {
            "volume": (0.05 - outflow_ - 0.01 * 0.4 + 0.4 * 0.5) * 1.5,
            "phosphate": 0,
            "temperature": temperature,
        }

        d6 = {
            "volume": outflow_ * 0.04 * 1.5 + 0.05 * 1.5,
            "phosphate": 0,
            "temperature": 10,
        }

        self.assertDictAlmostEqual(d1, r1)
        self.assertDictAlmostEqual(d2, r2)
        self.assertDictAlmostEqual(d3, surface.percolation, 15)
        self.assertDictAlmostEqual(d4, surface.subsurface_flow, 15)
        self.assertDictAlmostEqual(d5, surface.storage, 15)
        self.assertDictAlmostEqual(d6, surface.infiltration_excess)

    def test_perv_route(self):
        constants.set_simple_pollutants()
        land = Land(
            name="",
        )

        d1 = {"volume": 2.5, "temperature": 10, "phosphate": 0.3}

        d2 = {"volume": 2, "temperature": 12, "phosphate": 0.1}

        d3 = {"volume": 5, "temperature": 11, "phosphate": 0.2}

        d4 = {"volume": 0, "temperature": 0, "phosphate": 0}
        surface = PerviousSurface(parent=land, depth=0.5, area=1.5)
        surface.infiltration_excess = d1
        surface.subsurface_flow = d2
        surface.percolation = d3

        (r1, r2) = surface.route()
        self.assertDictAlmostEqual(d4, r1)
        self.assertDictAlmostEqual(d4, r2)
        self.assertDictAlmostEqual(d1, land.surface_runoff.storage)
        self.assertDictAlmostEqual(d2, land.subsurface_runoff.storage)
        self.assertDictAlmostEqual(d3, land.percolation.storage)

    def test_soil_temp(self):
        # Above field capacity
        constants.set_simple_pollutants()
        inputs = {
            ("temperature", 1): 10,
        }

        node = Node(name="", data_input_dict=inputs)
        node.t = 1
        surface = PerviousSurface(
            parent=node,
            depth=0.5,
            area=1.5,
            initial_storage={"volume": 7, "temperature": 3, "phosphate": 0.2},
        )

        surface.soil_temp_w_prev = 0.2
        surface.soil_temp_w_air = 0.3
        surface.soil_temp_w_deep = 0.4
        surface.soil_temp_deep = 5

        d1 = {"volume": 0, "temperature": 0, "phosphate": 0}

        d2 = {
            "volume": 7,
            "temperature": (5 * 0.4 + 0.2 * 3 + 0.3 * 10) / (0.2 + 0.3 + 0.4),
            "phosphate": 0.2,
        }

        (r1, r2) = surface.calculate_soil_temperature()
        self.assertDictAlmostEqual(d1, r1)
        self.assertDictAlmostEqual(d1, r2)
        self.assertDictAlmostEqual(d2, surface.storage, 14)

    def create_growing_surface(self):
        """

        Returns:

        """
        constants.set_default_pollutants()
        node = Node(name="")
        initial_vol = node.empty_vqip()
        initial_vol["phosphate"] = 11
        initial_vol["nitrate"] = 2.5
        initial_vol["nitrite"] = 1.5
        initial_vol["ammonia"] = 0.1
        initial_vol["org-nitrogen"] = 0.2
        initial_vol["org-phosphorus"] = 3
        initial_vol["volume"] = 0.32
        initial_vol["temperature"] = 11

        initial_soil = {
            "phosphate": 1.2,
            "ammonia": 0.2,
            "nitrate": 0.3,
            "nitrite": 0.4,
            "org-nitrogen": 2,
            "org-phosphorus": 4,
        }

        crop_factor_stages = [0.0, 0.0, 0.3, 0.3, 1.2, 1.2, 0.325, 0.0, 0.0]
        crop_factor_stage_dates = [0, 90, 91, 121, 161, 213, 244, 245, 366]
        sowing_day = 91
        harvest_day = 244
        ET_depletion_factor = 0.55

        surface = GrowingSurface(
            rooting_depth=0.5,
            area=1.5,
            initial_storage=initial_vol,
            initial_soil_storage=initial_soil,
            crop_factor_stages=crop_factor_stages,
            crop_factor_stage_dates=crop_factor_stage_dates,
            sowing_day=sowing_day,
            harvest_day=harvest_day,
            ET_depletion_factor=ET_depletion_factor,
            wilting_point=0.1,
            field_capacity=0.2,
        )
        return surface, initial_vol, initial_soil

    def test_grow_init(self):
        constants.set_default_pollutants()
        surface, ivol, isoil = self.create_growing_surface()

        d1 = {
            "N": ivol["nitrate"] + ivol["nitrite"] + ivol["ammonia"],
            "P": ivol["phosphate"],
        }
        self.assertDictAlmostEqual(
            surface.nutrient_pool.dissolved_inorganic_pool.storage, d1
        )

        d2 = {"N": ivol["org-nitrogen"], "P": ivol["org-phosphorus"]}
        self.assertDictAlmostEqual(
            surface.nutrient_pool.dissolved_organic_pool.storage, d2
        )

        d3 = {
            "N": isoil["nitrate"] + isoil["nitrite"] + isoil["ammonia"],
            "P": isoil["phosphate"],
        }
        self.assertDictAlmostEqual(
            surface.nutrient_pool.adsorbed_inorganic_pool.storage, d3, 15
        )

        d4 = {"N": isoil["org-nitrogen"], "P": isoil["org-phosphorus"]}
        self.assertDictAlmostEqual(surface.nutrient_pool.fast_pool.storage, d4)

    def test_grow_pull(self):
        constants.set_default_pollutants()
        surface, ivol, isoil = self.create_growing_surface()

        d1 = surface.empty_vqip()
        for key, amount in ivol.items():
            d1[key] = amount * 0.25 / 0.32
        d1["temperature"] = ivol["temperature"]
        n1 = {
            "N": (ivol["nitrate"] + ivol["nitrite"] + ivol["ammonia"])
            * (1 - 0.25 / 0.32),
            "P": ivol["phosphate"] * (1 - 0.25 / 0.32),
        }
        n2 = {
            "N": ivol["org-nitrogen"] * (1 - 0.25 / 0.32),
            "P": ivol["org-phosphorus"] * (1 - 0.25 / 0.32),
        }
        n3 = {
            "N": (isoil["nitrate"] + isoil["nitrite"] + isoil["ammonia"]),
            "P": isoil["phosphate"],
        }
        n4 = {"N": isoil["org-nitrogen"], "P": isoil["org-phosphorus"]}
        reply = surface.pull_storage({"volume": 0.25})
        self.assertDictAlmostEqual(d1, reply, 15)
        self.assertDictAlmostEqual(
            surface.nutrient_pool.dissolved_inorganic_pool.storage, n1, 15
        )
        self.assertDictAlmostEqual(
            surface.nutrient_pool.dissolved_organic_pool.storage, n2, 15
        )
        self.assertDictAlmostEqual(
            surface.nutrient_pool.adsorbed_inorganic_pool.storage, n3, 15
        )
        self.assertDictAlmostEqual(surface.nutrient_pool.fast_pool.storage, n4, 15)

    def test_crop_cover(self):
        constants.set_default_pollutants()
        surface, ivol, isoil = self.create_growing_surface()
        date = to_datetime("2000-05-01")
        node = Node(name="")
        node.t = date
        surface.parent = node
        surface.days_after_sow = date.dayofyear - surface.sowing_day
        _ = surface.calc_crop_cover()
        self.assertEqual(0.3, surface.crop_factor)
        self.assertEqual(0.3, surface.et0_coefficient)

        # These two numbers from the interpolation between
        # doy-harvest_sow_calendar and crop/ground_cover_stages
        self.assertEqual(0.17647058823529413, surface.crop_cover)
        self.assertEqual(0.058823529411764705, surface.ground_cover)

    def test_adjust_vqip(self):
        constants.set_default_pollutants()
        surface = GrowingSurface()
        vqip = surface.empty_vqip()
        vqip["nitrate"] = 0.5
        vqip["ammonia"] = 0.75
        vqip["org-nitrogen"] = 1.25
        vqip["phosphate"] = 1.2
        vqip["org-phosphorus"] = 0.3
        dep_ = {"N": 5, "P": 3}
        in_ = {"N": 3, "P": 2}

        d1 = surface.empty_vqip()
        d1["nitrate"] = 0.5 * 3 / 5
        d1["ammonia"] = 0.75 * 3 / 5
        d1["org-nitrogen"] = 1.25 * 3 / 5
        d1["phosphate"] = 1.2 * 2 / 3
        d1["org-phosphorus"] = 0.3 * 2 / 3

        r1 = surface.adjust_vqip_to_liquid(vqip, dep_, in_)
        self.assertDictAlmostEqual(d1, r1, 15)

    def test_dry_grow_dep(self):
        constants.set_default_pollutants()
        surface = GrowingSurface()
        vqip = surface.empty_vqip()
        vqip["nitrate"] = 0.5
        vqip["ammonia"] = 0.75
        vqip["phosphate"] = 1.2
        vqip_ = surface.dry_deposition_to_tank(vqip)

        d1 = surface.empty_vqip()
        d1["nitrate"] = 0.45
        d1["ammonia"] = 0.675
        d1["phosphate"] = 0
        self.assertDictAlmostEqual(d1, vqip_)
        n1 = {"N": 0, "P": 1.2}
        self.assertDictAlmostEqual(
            n1, surface.nutrient_pool.adsorbed_inorganic_pool.storage
        )

        n2 = {"N": (0.5 + 0.75) * surface.nutrient_pool.fraction_dry_n_to_fast, "P": 0}
        self.assertDictAlmostEqual(n2, surface.nutrient_pool.fast_pool.storage)

        n3 = {
            "N": (0.5 + 0.75)
            * surface.nutrient_pool.fraction_dry_n_to_dissolved_inorganic,
            "P": 0,
        }
        self.assertDictAlmostEqual(
            n3, surface.nutrient_pool.dissolved_inorganic_pool.storage
        )

    def test_wet_grow_dep(self):
        constants.set_default_pollutants()
        surface = GrowingSurface()
        vqip = surface.empty_vqip()
        vqip["nitrate"] = 0.5
        vqip["ammonia"] = 0.75
        vqip["phosphate"] = 1.2
        vqip_ = surface.wet_deposition_to_tank(vqip)

        d1 = surface.empty_vqip()
        d1["nitrate"] = 0.5
        d1["ammonia"] = 0.75
        d1["phosphate"] = 1.2
        self.assertDictAlmostEqual(d1, vqip_)

        n1 = {"N": (0.5 + 0.75), "P": 1.2}
        self.assertDictAlmostEqual(
            n1, surface.nutrient_pool.dissolved_inorganic_pool.storage
        )

    def test_fertiliser(self):
        constants.set_default_pollutants()
        surface, ivol, isoil = self.create_growing_surface()
        input_data = {
            ("nhx-fertiliser", 1): 5,
            ("noy-fertiliser", 1): 2,
            ("srp-fertiliser", 1): 3,
        }
        surface.data_input_dict = input_data
        node = Node(name="")
        node.monthyear = 1
        surface.parent = node

        (r1, r2) = surface.fertiliser()

        d1 = surface.empty_vqip()

        self.assertDictAlmostEqual(d1, r2)

        d1["phosphate"] = 3 * 1.5
        d1["nitrate"] = 2 * 1.5
        d1["ammonia"] = 5 * 1.5
        self.assertDictAlmostEqual(d1, r1)

        n1 = {"N": 4.1 + (2 + 5) * 1.5, "P": 11 + 3 * 1.5}
        self.assertDictAlmostEqual(
            n1, surface.nutrient_pool.dissolved_inorganic_pool.storage
        )

    def test_manure(self):
        constants.set_default_pollutants()
        surface, ivol, isoil = self.create_growing_surface()
        input_data = {("nhx-manure", 1): 5, ("noy-manure", 1): 2, ("srp-manure", 1): 3}
        surface.data_input_dict = input_data
        node = Node(name="")
        node.monthyear = 1
        surface.parent = node

        (r1, r2) = surface.manure()

        d1 = surface.empty_vqip()

        self.assertDictAlmostEqual(d1, r2)

        d1["phosphate"] = (
            3 * 1.5 * surface.nutrient_pool.fraction_manure_to_dissolved_inorganic["P"]
        )
        d1["nitrate"] = (
            2 * 1.5 * surface.nutrient_pool.fraction_manure_to_dissolved_inorganic["N"]
        )
        d1["ammonia"] = (
            5 * 1.5 * surface.nutrient_pool.fraction_manure_to_dissolved_inorganic["N"]
        )
        self.assertDictAlmostEqual(d1, r1)

        n1 = {
            "N": 4.1
            + (2 + 5)
            * 1.5
            * surface.nutrient_pool.fraction_manure_to_dissolved_inorganic["N"],
            "P": 11
            + 3
            * 1.5
            * surface.nutrient_pool.fraction_manure_to_dissolved_inorganic["P"],
        }
        self.assertDictAlmostEqual(
            n1, surface.nutrient_pool.dissolved_inorganic_pool.storage
        )

        n2 = {
            "N": 2 + (2 + 5) * 1.5 * surface.nutrient_pool.fraction_manure_to_fast["N"],
            "P": 4 + 3 * 1.5 * surface.nutrient_pool.fraction_manure_to_fast["P"],
        }
        self.assertDictAlmostEqual(n2, surface.nutrient_pool.fast_pool.storage)

    def calc_temp_dep(self):
        """"""
        constants.set_simple_pollutants()
        surface = GrowingSurface(
            initial_storage={"volume": 0, "phosphate": 0, "temperature": -1}
        )
        surface.calc_temperature_dependence_factor()
        self.assertEqual(0, surface.nutrient_pool.temperature_dependence_factor)

        surface.storage["temperature"] = 2.5
        surface.calc_temperature_dependence_factor()
        self.assertEqual(2.5 / 5, surface.nutrient_pool.temperature_dependence_factor)

        surface.storage["temperature"] = 7.5
        surface.calc_temperature_dependence_factor()
        self.assertEqual(
            2 ** ((7.5 - 20) / 10), surface.nutrient_pool.temperature_dependence_factor
        )

    def calc_mois_dep(self):
        """"""
        constants.set_simple_pollutants()
        surface = GrowingSurface(
            initial_storage={"volume": 0.05, "phosphate": 0, "temperature": 0},
            rooting_depth=1,
            area=1,
            wilting_point=0.1,
            field_capacity=0.2,
        )
        surface.calc_soil_moisture_dependence_factor()
        self.assertEqual(0, surface.nutrient_pool.soil_moisture_dependence_factor)

        surface.storage["volume"] = 0.15
        surface.calc_soil_moisture_dependence_factor()
        v1 = ((0.2 - 0.15) / (surface.thetaupp * 1)) ** surface.thetapow
        v1 = (1 - surface.satact) * v1 + surface.satact
        v2 = ((0.15 - 0.1) / (surface.thetalow * 1)) ** surface.thetapow

        v = min(1, v1, v2)
        self.assertEqual(v, surface.nutrient_pool.soil_moisture_dependence_factor)

        surface.storage["volume"] = 0.25
        surface.calc_soil_moisture_dependence_factor()
        self.assertEqual(
            surface.satact, surface.nutrient_pool.soil_moisture_dependence_factor
        )

    def test_soil_pool(self):
        constants.set_default_pollutants()
        surface, ivol, isoil = self.create_growing_surface()
        surface.calc_temperature_dependence_factor()
        surface.calc_soil_moisture_dependence_factor()

        nut_p = surface.nutrient_pool
        cf = nut_p.temperature_dependence_factor * nut_p.soil_moisture_dependence_factor
        (r1, r2) = surface.soil_pool_transformation()

        fast_ = {"N": isoil["org-nitrogen"], "P": isoil["org-phosphorus"]}  # original
        dissolved_inorganic_ = {
            "N": ivol["nitrate"] + ivol["nitrite"] + ivol["ammonia"],
            "P": ivol["phosphate"],
        }

        miner = {
            "N": nut_p.minfpar["N"] * fast_["N"] * cf,
            "P": nut_p.minfpar["P"] * fast_["P"] * cf,
        }

        fast_ = {"N": fast_["N"] - miner["N"], "P": fast_["P"] - miner["P"]}
        dissolved_inorganic_ = {
            "N": dissolved_inorganic_["N"] + miner["N"],
            "P": dissolved_inorganic_["P"] + miner["P"],
        }

        disso = {
            "N": nut_p.disfpar["N"] * fast_["N"] * cf,
            "P": nut_p.disfpar["P"] * fast_["P"] * cf,
        }

        fast_ = {"N": fast_["N"] - disso["N"], "P": fast_["P"] - disso["P"]}

        immob = {
            "N": nut_p.immobdpar["N"] * dissolved_inorganic_["N"] * cf,
            "P": nut_p.immobdpar["P"] * dissolved_inorganic_["P"] * cf,
        }

        d1 = surface.empty_vqip()
        d2 = surface.empty_vqip()
        if miner["P"] - immob["P"] > 0:
            d1["phosphate"] = miner["P"] - immob["P"]
        else:
            d2["phosphate"] = immob["P"] - miner["P"]
        if miner["N"] - immob["N"] > 0:
            d1["nitrate"] = (
                (miner["N"] - immob["N"])
                * ivol["nitrate"]
                / (ivol["nitrate"] + ivol["ammonia"])
            )
            d1["ammonia"] = (
                (miner["N"] - immob["N"])
                * ivol["ammonia"]
                / (ivol["nitrate"] + ivol["ammonia"])
            )
        else:
            d2["nitrate"] = (
                (immob["N"] - miner["N"])
                * ivol["nitrate"]
                / (ivol["nitrate"] + ivol["ammonia"])
            )
            d2["ammonia"] = (
                (immob["N"] - miner["N"])
                * ivol["ammonia"]
                / (ivol["nitrate"] + ivol["ammonia"])
            )
        d1["org-phosphorus"] = disso["P"]
        d1["org-nitrogen"] = disso["N"]

        self.assertDictAlmostEqual(d1, r1, 14)
        self.assertDictAlmostEqual(d2, r2, 15)

    def test_crop_uptake(self):
        constants.set_default_pollutants()
        surface, ivol, isoil = self.create_growing_surface()
        date = to_datetime("2000-05-01")
        node = Node(name="")
        node.t = date
        surface.parent = node
        surface.days_after_sow = date.dayofyear - surface.sowing_day
        _ = surface.calc_crop_cover()

        (r1, r2) = surface.calc_crop_uptake()

        d1 = surface.empty_vqip()
        self.assertDictAlmostEqual(r1, d1)
        d1["nitrate"] = 4.7281041425055e-05
        d1["phosphate"] = 6.5668113090354e-06
        self.assertDictAlmostEqual(d1, r2)

        d2 = surface.copy_vqip(ivol)
        d2["nitrate"] -= d1["nitrate"]
        d2["phosphate"] -= d1["phosphate"]
        self.assertDictAlmostEqual(d2, surface.storage)

        n1 = {
            "N": ivol["nitrate"] + ivol["ammonia"] + ivol["nitrite"] - d1["nitrate"],
            "P": ivol["phosphate"] - d1["phosphate"],
        }
        self.assertDictAlmostEqual(
            n1, surface.nutrient_pool.dissolved_inorganic_pool.storage
        )

    def test_erosion1(self):
        # High rain
        constants.set_default_pollutants()
        surface, ivol, isoil = self.create_growing_surface()

        date = to_datetime("2000-05-01")
        inputs = {
            ("precipitation", date): 1.5,
            ("et0", date): 0.01,
            ("temperature", date): 10,
        }

        node = Node(name="", data_input_dict=inputs)
        node.t = date
        surface.parent = node
        _ = surface.ihacres()
        (r1, r2) = surface.erosion()
        d1 = surface.empty_vqip()
        self.assertDictAlmostEqual(d1, r2)
        d1["phosphate"] = 0.0019497135091285024
        d1["solids"] = 1.0560948174446056
        self.assertDictAlmostEqual(d1, r1)

    def test_erosion2(self):
        # Med rain
        constants.set_default_pollutants()
        surface, ivol, isoil = self.create_growing_surface()

        date = to_datetime("2000-05-01")
        inputs = {
            ("precipitation", date): 0.1,
            ("et0", date): 0.01,
            ("temperature", date): 10,
        }

        node = Node(name="", data_input_dict=inputs)
        node.t = date
        surface.parent = node
        _ = surface.ihacres()
        (r1, r2) = surface.erosion()
        d1 = surface.empty_vqip()
        self.assertDictAlmostEqual(d1, r2)
        d1["phosphate"] = 1.0244298083948e-06
        d1["solids"] = 0.0005548994795471758
        self.assertDictAlmostEqual(d1, r1)

    def test_erosion3(self):
        # low rain
        constants.set_default_pollutants()
        surface, ivol, isoil = self.create_growing_surface()

        date = to_datetime("2000-05-01")
        inputs = {
            ("precipitation", date): 0.005,
            ("et0", date): 0.01,
            ("temperature", date): 10,
        }

        node = Node(name="", data_input_dict=inputs)
        node.t = date
        surface.parent = node
        _ = surface.ihacres()
        (r1, r2) = surface.erosion()
        d1 = surface.empty_vqip()
        self.assertDictAlmostEqual(d1, r2)
        d1["phosphate"] = 8.3285749289e-09
        d1["solids"] = 1.7605117735786e-06
        self.assertDictAlmostEqual(d1, r1)

    def test_denitrification(self):
        # low rain
        constants.set_default_pollutants()
        surface, ivol, isoil = self.create_growing_surface()
        surface.calc_temperature_dependence_factor()
        (r1, r2) = surface.denitrification()
        d1 = surface.empty_vqip()
        self.assertDictAlmostEqual(d1, r1)
        d1["nitrate"] = 0.03295446191742672
        self.assertDictAlmostEqual(d1, r2)

    def test_desorption(self):
        constants.set_default_pollutants()
        surface, ivol, isoil = self.create_growing_surface()
        surface.nutrient_pool.adsorbed_inorganic_pool.storage["P"] = 1e6
        (r1, r2) = surface.adsorption()
        d1 = surface.empty_vqip()
        self.assertDictAlmostEqual(d1, r2)
        d1["phosphate"] = 29534.602916697728
        self.assertDictAlmostEqual(d1, r1)
        n1 = {"N": 0.9, "P": 970465.3970833023}
        self.assertDictAlmostEqual(
            n1, surface.nutrient_pool.adsorbed_inorganic_pool.storage
        )
        n2 = {"N": 4.1, "P": 29545.602916697728}
        self.assertDictAlmostEqual(
            n2, surface.nutrient_pool.dissolved_inorganic_pool.storage
        )

    def test_adsorption(self):
        constants.set_default_pollutants()
        surface, ivol, isoil = self.create_growing_surface()

        input_data = {
            ("nhx-fertiliser", 1): 0,
            ("noy-fertiliser", 1): 0,
            ("srp-fertiliser", 1): 1000,
        }
        surface.data_input_dict = input_data
        node = Node(name="")
        node.monthyear = 1
        surface.parent = node

        surface.fertiliser()

        (r1, r2) = surface.adsorption()
        d1 = surface.empty_vqip()
        self.assertDictAlmostEqual(d1, r1)
        d1["phosphate"] = 1.5761692429741743
        self.assertDictAlmostEqual(d1, r2)
        n1 = {"N": 0.9, "P": 2.7761692429741744}
        self.assertDictAlmostEqual(
            n1, surface.nutrient_pool.adsorbed_inorganic_pool.storage
        )
        n2 = {"N": 4.1, "P": 1509.4238307570258}
        self.assertDictAlmostEqual(
            n2, surface.nutrient_pool.dissolved_inorganic_pool.storage
        )

    def test_irrigate(self):
        constants.set_default_pollutants()

        date = to_datetime("2000-05-01")
        inputs = {
            ("precipitation", date): 0.005,
            ("et0", date): 0.01,
            ("temperature", date): 10,
        }

        node = Node(name="", data_input_dict=inputs)
        node.t = date

        initial_vol = node.empty_vqip()
        initial_vol["phosphate"] = 11
        initial_vol["nitrate"] = 2.5
        initial_vol["nitrite"] = 1.5
        initial_vol["ammonia"] = 0.1
        initial_vol["org-nitrogen"] = 0.2
        initial_vol["org-phosphorus"] = 3
        initial_vol["volume"] = 0.3
        initial_vol["temperature"] = 11

        initial_soil = {
            "phosphate": 1.2,
            "ammonia": 0.2,
            "nitrate": 0.3,
            "nitrite": 0.4,
            "org-nitrogen": 2,
            "org-phosphorus": 4,
        }

        crop_factor_stages = [0.0, 0.0, 0.3, 0.3, 1.2, 1.2, 0.325, 0.0, 0.0]
        crop_factor_stage_dates = [0, 90, 91, 121, 161, 213, 244, 245, 366]
        sowing_day = 91
        harvest_day = 244
        ET_depletion_factor = 0.55

        surface = IrrigationSurface(
            rooting_depth=0.5,
            area=1.5,
            initial_storage=initial_vol,
            initial_soil_storage=initial_soil,
            crop_factor_stages=crop_factor_stages,
            crop_factor_stage_dates=crop_factor_stage_dates,
            sowing_day=sowing_day,
            harvest_day=harvest_day,
            ET_depletion_factor=ET_depletion_factor,
            wilting_point=0.1,
            field_capacity=0.2,
            total_porosity=0.4,
            irrigation_coefficient=0.8,
            parent=node,
        )
        surface.days_after_sow = date.dayofyear - surface.sowing_day
        _ = surface.ihacres()

        reservoir = Reservoir(name="", capacity=50, initial_storage=40)
        arc = Arc(in_port=reservoir, out_port=node, name="")
        surface.irrigate()
        self.assertEqual(0.006, arc.flow_in)

    def test_garden(self):
        constants.set_simple_pollutants()

        date = to_datetime("2000-05-01")
        inputs = {
            ("precipitation", date): 0.0025,
            ("et0", date): 0.01,
            ("temperature", date): 10,
        }

        node = Node(name="", data_input_dict=inputs)
        node.t = date

        surface = GardenSurface(
            rooting_depth=0.5,
            area=1.5,
            wilting_point=0.1,
            field_capacity=0.2,
            initial_storage=0.2 * 0.5 * 1.5,
            parent=node,
        )

        _ = surface.ihacres()

        reply = surface.calculate_irrigation_demand()
        d1 = {"phosphate": 0, "temperature": 0, "volume": 0.0075 * 1.5}
        self.assertDictAlmostEqual(d1, reply)

        d2 = {
            "phosphate": 0,
            "temperature": 0,
            "volume": 0.2 * 0.5 * 1.5 - 0.0075 * 1.5,
        }
        self.assertDictAlmostEqual(d2, surface.storage, 16)

        reply = surface.receive_irrigation_demand(d1)
        self.assertDictAlmostEqual(surface.empty_vqip(), reply)

        d3 = {"phosphate": 0, "temperature": 0, "volume": 0.2 * 0.5 * 1.5}
        self.assertDictAlmostEqual(d3, surface.storage, 16)

    def test_land_overrides(self):
        constants.set_simple_pollutants()
        node = Land(name="")
        node.apply_overrides(
            {
                "surface_residence_time": 4.9,
                "subsurface_residence_time": 23.7,
                "percolation_residence_time": 56.1,
            }
        )
        self.assertEqual(node.surface_residence_time, 4.9)
        self.assertEqual(node.surface_runoff.residence_time, 4.9)
        self.assertEqual(node.subsurface_residence_time, 23.7)
        self.assertEqual(node.subsurface_runoff.residence_time, 23.7)
        self.assertEqual(node.percolation_residence_time, 56.1)

    def test_surface_overrides(self):
        constants.set_default_pollutants()
        decaytank = DecayTank()
        surface = Surface(
            parent=decaytank, area=5, depth=0.1, pollutant_load={"nitrate": 5.7}
        )
        surface.apply_overrides(
            {
                "surface": "test_surface",
                "area": 9.8,
                "depth": 7.5,
                "pollutant_load": {"phosphate": 10.1},
                "decays": {"nitrate": {"constant": 0.001, "exponent": 1.005}},
            }
        )
        self.assertEqual(surface.surface, "test_surface")
        self.assertEqual(surface.area, 9.8)
        self.assertEqual(surface.depth, 7.5)
        self.assertEqual(surface.capacity, 9.8 * 7.5)
        self.assertDictEqual(
            surface.pollutant_load, {"nitrate": 5.7, "phosphate": 10.1}
        )
        self.assertDictEqual(
            surface.decays, {"nitrate": {"constant": 0.001, "exponent": 1.005}}
        )

        # override the data_input_dict
        # test the format of dict
        new_data_input_dict = {
            ("temperature", 1): 10,
            ("temperature", 2): 20,
        }
        surface.apply_overrides({"data_input_dict": new_data_input_dict})
        self.assertDictEqual(surface.data_input_dict, new_data_input_dict)
        # test the format of str
        new_data_input_dict = str(
            Path(__file__).parent / "example_data_input_dict.csv.gz"
        )
        surface.apply_overrides({"data_input_dict": new_data_input_dict})
        from wsimod.orchestration.model import read_csv

        new_data_input_dict = read_csv(new_data_input_dict)
        self.assertDictEqual(surface.data_input_dict, new_data_input_dict)
        print(dict(list(surface.data_input_dict.items())[:5]))

    def test_impervioussurface_overrides(self):
        constants.set_default_pollutants()
        decaytank = DecayTank()
        surface = Surface(parent=decaytank, area=5, depth=0.1)
        impervioussurface = ImperviousSurface(
            parent=surface, area=1.5, et0_to_e=0.9, pore_depth=0.015
        )
        impervioussurface.apply_overrides(
            {"surface": "test_surface", "area": 9.8, "pore_depth": 7.5, "et0_to_e": 3.5}
        )
        self.assertEqual(impervioussurface.area, 9.8)
        self.assertEqual(impervioussurface.pore_depth, 7.5)
        self.assertEqual(impervioussurface.depth, 7.5)
        self.assertEqual(impervioussurface.capacity, 9.8 * 7.5)
        self.assertEqual(impervioussurface.et0_to_e, 3.5)

    def test_pervioussurface_overrides(self):
        constants.set_default_pollutants()
        decaytank = DecayTank()
        surface = Surface(parent=decaytank, area=5, depth=0.1)
        pervioussurface = PerviousSurface(
            parent=surface, depth=0.5, area=1.5, initial_storage=0.5 * 1.5 * 0.25
        )
        pervioussurface.apply_overrides(
            {
                "field_capacity": 0.335,
                "wilting_point": 0.112,
                "total_porosity": 0.476,
                "infiltration_capacity": 0.678,
                "surface_coefficient": 0.237,
                "percolation_coefficient": 0.777,
                "et0_coefficient": 0.697,
                "ihacres_p": 10.096,
                "soil_temp_w_prev": 37.1,
                "soil_temp_w_air": 23.6,
                "soil_temp_w_deep": 3.4,
                "soil_temp_deep": 2.2,
                "surface": "test_surface",
                "area": 9.8,
                "depth": 7.5,
            }
        )
        self.assertEqual(pervioussurface.field_capacity, 0.335)
        self.assertEqual(pervioussurface.wilting_point, 0.112)
        self.assertEqual(pervioussurface.total_porosity, 0.476)
        self.assertEqual(pervioussurface.infiltration_capacity, 0.678)
        self.assertEqual(pervioussurface.surface_coefficient, 0.237)
        self.assertEqual(pervioussurface.percolation_coefficient, 0.777)
        self.assertEqual(pervioussurface.et0_coefficient, 0.697)
        self.assertEqual(pervioussurface.ihacres_p, 10.096)
        self.assertEqual(pervioussurface.soil_temp_w_prev, 37.1)
        self.assertEqual(pervioussurface.soil_temp_w_air, 23.6)
        self.assertEqual(pervioussurface.soil_temp_w_deep, 3.4)
        self.assertEqual(pervioussurface.soil_temp_deep, 2.2)

        self.assertEqual(pervioussurface.field_capacity_m, 0.335 * 7.5)
        self.assertEqual(pervioussurface.wilting_point_m, 0.112 * 7.5)
        self.assertEqual(pervioussurface.depth, 0.476 * 7.5)
        self.assertEqual(pervioussurface.area, 9.8)
        self.assertEqual(pervioussurface.capacity, 7.5 * 0.476 * 9.8)
        self.assertEqual(pervioussurface.surface, "test_surface")
        self.assertEqual(pervioussurface.subsurface_coefficient, 1 - 0.777)

    def test_growingsurface_overrides(self):
        constants.set_default_pollutants()
        decaytank = DecayTank()
        surface = Surface(parent=decaytank, area=5, depth=0.1)
        pervioussurface = PerviousSurface(
            parent=surface, depth=0.5, area=1.5, initial_storage=0.5 * 1.5 * 0.25
        )
        growingsurface = GrowingSurface(parent=pervioussurface, area=1.5)
        overrides = {
            "ET_depletion_factor": 0.521,
            "crop_cover_max": 1.342,
            "ground_cover_max": 1.111,
            "crop_factor_stages": [1, 2, 1],
            "crop_factor_stage_dates": [1, 32, 90],
            "sowing_day": 35,
            "harvest_day": 89,
            "satact": 0.567,
            "thetaupp": 4.324,
            "thetalow": 3.582,
            "thetapow": 7.324,
            "uptake1": 1.278,
            "uptake2": 2.753,
            "uptake3": 5.298,
            "uptake_PNratio": 3.263,
            "erodibility": 2.863,
            "sreroexp": 5.634,
            "cohesion": 8.903,
            "slope": 6.231,
            "srfilt": 9.231,
            "macrofilt": 7.394,
            "limpar": 4.211,
            "exppar": 5.872,
            "hsatINs": 20.321,
            "denpar": 0.204,
            "adosorption_nr_limit": 1.943,
            "adsorption_nr_maxiter": 6321,
            "kfr": 80.2,
            "nfr": 42.3,
            "kadsdes": 0.972,
            "bulk_density": 1672,
            "field_capacity": 0.335,
            "wilting_point": 0.112,
            "total_porosity": 0.476,
            "rooting_depth": 7.5,
        }
        overrides_to_check = overrides.copy()
        growingsurface.apply_overrides(overrides)

        for k, v in overrides_to_check.items():
            if isinstance(v, list):
                self.assertListEqual(getattr(growingsurface, k), v)
            else:
                self.assertEqual(getattr(growingsurface, k), v)

        harvest_sow_calendar = [
            0,
            35,
            89,
            89 + 1,
            365,
        ]
        self.assertListEqual(growingsurface.harvest_sow_calendar, harvest_sow_calendar)
        self.assertListEqual(growingsurface.ground_cover_stages, [0, 0, 1.111, 0, 0])
        self.assertListEqual(growingsurface.crop_cover_stages, [0, 0, 1.342, 0, 0])
        self.assertEqual(growingsurface.autumn_sow, False)
        self.assertEqual(growingsurface.total_available_water, (0.335 - 0.112) * 7.5)
        self.assertEqual(
            growingsurface.readily_available_water, (0.335 - 0.112) * 7.5 * 0.521
        )
        self.assertEqual(growingsurface.depth, 0.476 * 7.5)
        self.assertEqual(growingsurface.capacity, 0.476 * 7.5 * 1.5)

    def test_nutrientpool_overrides(self):
        constants.set_default_pollutants()
        decaytank = DecayTank()
        surface = Surface(parent=decaytank, area=5, depth=0.1)
        pervioussurface = PerviousSurface(
            parent=surface, depth=0.5, area=1.5, initial_storage=0.5 * 1.5 * 0.25
        )
        growingsurface = GrowingSurface(parent=pervioussurface, area=1.5)
        overrides = {
            "fraction_dry_n_to_dissolved_inorganic": 0.29,
            "degrhpar": {"N": 7 * 1e-1},
            "dishpar": {"P": 7 * 1e-1},
            "minfpar": {"N": 1.00013, "P": 1.000003},
            "disfpar": {"N": 1.000003, "P": 1.0000001},
            "immobdpar": {"N": 1.0056, "P": 1.2866},
            "fraction_manure_to_dissolved_inorganic": {"N": 0.35, "P": 0.21},
            "fraction_residue_to_fast": {"N": 0.61, "P": 0.71},
        }
        growingsurface.nutrient_pool.apply_overrides(overrides)
        self.assertEqual(
            growingsurface.nutrient_pool.fraction_dry_n_to_dissolved_inorganic, 0.29
        )
        self.assertDictEqual(
            growingsurface.nutrient_pool.degrhpar, {"N": 7 * 1e-1, "P": 7 * 1e-6}
        )
        self.assertDictEqual(
            growingsurface.nutrient_pool.dishpar, {"N": 7 * 1e-5, "P": 7 * 1e-1}
        )
        self.assertDictEqual(
            growingsurface.nutrient_pool.minfpar, {"N": 1.00013, "P": 1.000003}
        )
        self.assertDictEqual(
            growingsurface.nutrient_pool.disfpar, {"N": 1.000003, "P": 1.0000001}
        )
        self.assertDictEqual(
            growingsurface.nutrient_pool.immobdpar, {"N": 1.0056, "P": 1.2866}
        )
        self.assertDictEqual(
            growingsurface.nutrient_pool.fraction_manure_to_dissolved_inorganic,
            {"N": 0.35, "P": 0.21},
        )
        self.assertDictEqual(
            growingsurface.nutrient_pool.fraction_residue_to_fast,
            {"N": 0.61, "P": 0.71},
        )

        self.assertDictEqual(
            growingsurface.nutrient_pool.fraction_manure_to_fast,
            {"N": 1 - 0.35, "P": 1 - 0.21},
        )
        self.assertDictEqual(
            growingsurface.nutrient_pool.fraction_residue_to_humus,
            {"N": 1 - 0.61, "P": 1 - 0.71},
        )
        self.assertEqual(growingsurface.nutrient_pool.fraction_dry_n_to_fast, 0.71)


if __name__ == "__main__":
    unittest.main()



"""Module for defining tanks."""

from typing import Any, Dict

from wsimod.arcs.arcs import AltQueueArc, DecayArcAlt
from wsimod.core import constants
from wsimod.core.core import DecayObj, WSIObj


class Tank(WSIObj):
    """"""

    def __init__(self, capacity=0, area=1, datum=10, initial_storage=0):
        """A standard storage object.

        Args:
            capacity (float, optional): Volumetric tank capacity. Defaults to 0.
            area (float, optional): Area of tank. Defaults to 1.
            datum (float, optional): Datum of tank base (not currently used in any
                functions). Defaults to 10.
            initial_storage (optional): Initial storage for tank.
                float: Tank will be initialised with zero pollutants and the float
                    as volume
                dict: Tank will be initialised with this VQIP
                Defaults to 0 (i.e., no volume, no pollutants).
        """
        # Set parameters
        self.capacity = capacity
        self.area = area
        self.datum = datum
        self.initial_storage = initial_storage

        WSIObj.__init__(self)  # Not sure why I do this rather than super()

        # TODO I don't think the outer if statement is needed
        if "initial_storage" in dir(self):
            if isinstance(self.initial_storage, dict):
                # Assume dict is VQIP describing storage
                self.storage = self.copy_vqip(self.initial_storage)
                self.storage_ = self.copy_vqip(
                    self.initial_storage
                )  # Lagged storage for mass balance
            else:
                # Assume number describes initial stroage
                self.storage = self.v_change_vqip(
                    self.empty_vqip(), self.initial_storage
                )
                self.storage_ = self.v_change_vqip(
                    self.empty_vqip(), self.initial_storage
                )  # Lagged storage for mass balance
        else:
            self.storage = self.empty_vqip()
            self.storage_ = self.empty_vqip()  # Lagged storage for mass balance

    def apply_overrides(self, overrides: Dict[str, Any] = {}):
        """Apply overrides to the tank.

        Enables a user to override any of the following parameters:
        area, capacity, datum.

        Args:
            overrides (dict, optional): Dictionary of overrides. Defaults to {}.
        """
        self.capacity = overrides.pop("capacity", self.capacity)
        self.area = overrides.pop("area", self.area)
        self.datum = overrides.pop("datum", self.datum)
        if len(overrides) > 0:
            print(f"No override behaviour defined for: {overrides.keys()}")

    def ds(self):
        """Should be called by parent object to get change in storage.

        Returns:
            (dict): Change in storage
        """
        return self.ds_vqip(self.storage, self.storage_)

    def pull_ponded(self):
        """Pull any volume that is above the tank's capacity.

        Returns:
            ponded (vqip): Amount of ponded water that has been removed from the
                tank

        Examples:
            >>> constants.ADDITIVE_POLLUTANTS = ['phosphate']
            >>> my_tank = Tank(capacity = 9, initial_storage = {'volume' : 10,
                'phosphate' : 0.2})
            >>> print(my_tank.storage)
            {'volume' : 10, 'phosphate' : 0.2}
            >>> print(my_tank.pull_ponded())
            {'volume' : 1, 'phosphate' : 0.02}
            >>> print(my_tank.storage)
            {'volume' : 9, 'phosphate' : 0.18}
        """
        # Get amount
        ponded = max(self.storage["volume"] - self.capacity, 0)
        # Pull from tank
        ponded = self.pull_storage({"volume": ponded})
        return ponded

    def get_avail(self, vqip=None):
        """Get minimum of the amount of water in storage and vqip (if provided).

        Args:
            vqip (dict, optional): Maximum water required (only 'volume' is used).
                Defaults to None.

        Returns:
            reply (dict): Water available

        Examples:
            >>> constants.ADDITIVE_POLLUTANTS = ['phosphate']
            >>> my_tank = Tank(capacity = 9, initial_storage = {'volume' : 10,
                'phosphate' : 0.2})
            >>> print(my_tank.storage)
            {'volume' : 10, 'phosphate' : 0.2}
            >>> print(my_tank.get_avail())
            {'volume' : 10, 'phosphate' : 0.2}
            >>> print(my_tank.get_avail({'volume' : 1}))
            {'volume' : 1, 'phosphate' : 0.02}
        """
        reply = self.copy_vqip(self.storage)
        if vqip is None:
            # Return storage
            return reply
        else:
            # Adjust storage pollutants to match volume in vqip
            reply = self.v_change_vqip(reply, min(reply["volume"], vqip["volume"]))
            return reply

    def get_excess(self, vqip=None):
        """Get difference between current storage and tank capacity.

        Args:
            vqip (dict, optional): Maximum capacity required (only 'volume' is
                used). Defaults to None.

        Returns:
            (dict): Difference available

        Examples:
            >>> constants.ADDITIVE_POLLUTANTS = ['phosphate']
            >>> my_tank = Tank(capacity = 9, initial_storage = {'volume' : 5,
                'phosphate' : 0.2})
            >>> print(my_tank.get_excess())
            {'volume' : 4, 'phosphate' : 0.16}
            >>> print(my_tank.get_excess({'volume' : 2}))
            {'volume' : 2, 'phosphate' : 0.08}
        """
        vol = max(self.capacity - self.storage["volume"], 0)
        if vqip is not None:
            vol = min(vqip["volume"], vol)

        # Adjust storage pollutants to match volume in vqip
        # TODO the v_change_vqip in the reply here is a weird default (if a VQIP is not
        #   provided)
        return self.v_change_vqip(self.storage, vol)

    def push_storage(self, vqip, force=False):
        """Push water into tank, updating the storage VQIP. Force argument can be used
        to ignore tank capacity.

        Args:
            vqip (dict): VQIP amount to be pushed
            force (bool, optional): Argument used to cause function to ignore tank
                capacity, possibly resulting in pooling. Defaults to False.

        Returns:
            reply (dict): A VQIP of water not successfully pushed to the tank

        Examples:
            >>> constants.ADDITIVE_POLLUTANTS = ['phosphate']
            >>> constants.POLLUTANTS = ['phosphate']
            >>> constants.NON_ADDITIVE_POLLUTANTS = []
            >>> my_tank = Tank(capacity = 9, initial_storage = {'volume' : 5,
                'phosphate' : 0.2})
            >>> my_push = {'volume' : 10, 'phosphate' : 0.5}
            >>> reply = my_tank.push_storage(my_push)
            >>> print(reply)
            {'volume' : 6, 'phosphate' : 0.3}
            >>> print(my_tank.storage)
            {'volume': 9.0, 'phosphate': 0.4}
            >>> print(my_tank.push_storage(reply, force = True))
            {'phosphate': 0, 'volume': 0}
            >>> print(my_tank.storage)
            {'volume': 15.0, 'phosphate': 0.7}
        """
        if force:
            # Directly add request to storage
            self.storage = self.sum_vqip(self.storage, vqip)
            return self.empty_vqip()

        # Check whether request can be met
        excess = self.get_excess()["volume"]

        # Adjust accordingly
        reply = max(vqip["volume"] - excess, 0)
        reply = self.v_change_vqip(vqip, reply)
        entered = self.v_change_vqip(vqip, vqip["volume"] - reply["volume"])

        # Update storage
        self.storage = self.sum_vqip(self.storage, entered)

        return reply

    def pull_storage(self, vqip):
        """Pull water from tank, updating the storage VQIP. Pollutants are removed from
        tank in proportion to 'volume' in vqip (pollutant values in vqip are ignored).

        Args:
            vqip (dict): VQIP amount to be pulled, (only 'volume' key is needed)

        Returns:
            reply (dict): A VQIP water successfully pulled from the tank

        Examples:
            >>> constants.ADDITIVE_POLLUTANTS = ['phosphate']
            >>> my_tank = Tank(capacity = 9, initial_storage = {'volume' : 5,
                'phosphate' : 0.2})
            >>> print(my_tank.pull_storage({'volume' : 6}))
            {'volume': 5.0, 'phosphate': 0.2}
            >>> print(my_tank.storage)
            {'volume': 0, 'phosphate': 0}
        """
        # Pull from Tank by volume (taking pollutants in proportion)
        if self.storage["volume"] == 0:
            return self.empty_vqip()

        # Adjust based on available volume
        reply = min(vqip["volume"], self.storage["volume"])

        # Update reply to vqip (in proportion to concentration in storage)
        reply = self.v_change_vqip(self.storage, reply)

        # Extract from storage
        self.storage = self.extract_vqip(self.storage, reply)

        return reply

    def pull_pollutants(self, vqip):
        """Pull water from tank, updating the storage VQIP. Pollutants are removed from
        tank in according to their values in vqip.

        Args:
            vqip (dict): VQIP amount to be pulled

        Returns:
            vqip (dict): A VQIP water successfully pulled from the tank

        Examples:
            >>> constants.ADDITIVE_POLLUTANTS = ['phosphate']
            >>> my_tank = Tank(capacity = 9, initial_storage = {'volume' : 5,
                'phosphate' : 0.2})
            >>> print(my_tank.pull_pollutants({'volume' : 2, 'phosphate' : 0.15}))
            {'volume': 2.0, 'phosphate': 0.15}
            >>> print(my_tank.storage)
            {'volume': 3, 'phosphate': 0.05}
        """
        # Adjust based on available mass
        for pol in constants.ADDITIVE_POLLUTANTS + ["volume"]:
            vqip[pol] = min(self.storage[pol], vqip[pol])

        # Extract from storage
        self.storage = self.extract_vqip(self.storage, vqip)
        return vqip

    def get_head(self, datum=None, non_head_storage=0):
        """Area volume calculation for head calcuations. Datum and storage that does not
        contribute to head can be specified.

        Args:
            datum (float, optional): Value to add to pressure head in tank.
                Defaults to None.
            non_head_storage (float, optional): Amount of storage that does
                not contribute to generation of head. The tank must exceed
                this value to generate any pressure head. Defaults to 0.

        Returns:
            head (float): Total head in tank

        Examples:
            >>> my_tank = Tank(datum = 10, initial_storage = 5, capacity = 10, area = 2)
            >>> print(my_tank.get_head())
            12.5
            >>> print(my_tank.get_head(non_head_storage = 1))
            12
            >>> print(my_tank.get_head(non_head_storage = 1, datum = 0))
            2
        """
        # If datum not provided use object datum
        if datum is None:
            datum = self.datum

        # Calculate pressure head generating storage
        head_storage = max(self.storage["volume"] - non_head_storage, 0)

        # Perform head calculation
        head = head_storage / self.area + datum

        return head

    def evaporate(self, evap):
        """Wrapper for v_distill_vqip to apply a volumetric subtraction from tank
        storage. Volume removed from storage and no change in pollutant values.

        Args:
            evap (float): Volume to evaporate

        Returns:
            evap (float): Volumetric amount of evaporation successfully removed
        """
        avail = self.get_avail()["volume"]

        evap = min(evap, avail)
        self.storage = self.v_distill_vqip(self.storage, evap)
        return evap

    ##Old function no longer needed (check it is not used anywhere and remove)
    def push_total(self, vqip):
        """

        Args:
            vqip:

        Returns:

        """
        self.storage = self.sum_vqip(self.storage, vqip)
        return self.empty_vqip()

    ##Old function no longer needed (check it is not used anywhere and remove)
    def push_total_c(self, vqip):
        """

        Args:
            vqip:

        Returns:

        """
        # Push vqip to storage where pollutants are given as a concentration rather
        #   than storage
        vqip = self.concentration_to_total(self.vqip)
        self.storage = self.sum_vqip(self.storage, vqip)
        return self.empty_vqip()

    def end_timestep(self):
        """Function to be called by parent object, tracks previously timestep's
        storage."""
        self.storage_ = self.copy_vqip(self.storage)

    def reinit(self):
        """Set storage to an empty VQIP."""
        self.storage = self.empty_vqip()
        self.storage_ = self.empty_vqip()


class ResidenceTank(Tank):
    """"""

    def __init__(self, residence_time=2, **kwargs):
        """A tank that has a residence time property that limits storage pulled from the
        'pull_outflow' function.

        Args:
            residence_time (float, optional): Residence time, in theory given
                in timesteps, in practice it just means that storage /
                residence time can be pulled each time pull_outflow is called.
                Defaults to 2.
        """
        self.residence_time = residence_time
        super().__init__(**kwargs)

    def apply_overrides(self, overrides: Dict[str, Any] = {}):
        """Apply overrides to the residencetank.

        Enables a user to override any of the following parameters:
        residence_time.

        Args:
            overrides (dict, optional): Dictionary of overrides. Defaults to {}.
        """
        self.residence_time = overrides.pop("residence_time", self.residence_time)
        super().apply_overrides(overrides)

    def pull_outflow(self):
        """Pull storage by residence time from the tank, updating tank storage.

        Returns:
            outflow (dict): A VQIP with volume of pulled volume and pollutants
                proportionate to the tank's pollutants
        """
        # Calculate outflow
        outflow = self.storage["volume"] / self.residence_time
        # Update pollutant amounts
        outflow = self.v_change_vqip(self.storage, outflow)
        # Remove from tank
        outflow = self.pull_storage(outflow)
        return outflow


class DecayTank(Tank, DecayObj):
    """"""

    def __init__(self, decays={}, parent=None, **kwargs):
        """A tank that has DecayObj functions. Decay occurs in end_timestep, after
        updating state variables. In this sense, decay is occurring at the very
        beginning of the timestep.

        Args:
            decays (dict): A dict of dicts containing a key for each pollutant that
                decays and, within that, a key for each parameter (a constant and
                exponent)
            parent (object): An object that can be used to read temperature data from
        """
        # Store parameters
        self.parent = parent

        # Initialise Tank
        Tank.__init__(self, **kwargs)

        # Initialise decay object
        DecayObj.__init__(self, decays)

        # Update timestep and ds functions
        self.end_timestep = self.end_timestep_decay
        self.ds = self.decay_ds

    def apply_overrides(self, overrides: Dict[str, Any] = {}):
        """Apply overrides to the decaytank.

        Enables a user to override any of the following parameters:
        decays.

        Args:
            overrides (dict, optional): Dictionary of overrides. Defaults to {}.
        """
        self.decays.update(overrides.pop("decays", {}))
        super().apply_overrides(overrides)

    def end_timestep_decay(self):
        """Update state variables and call make_decay."""
        self.total_decayed = self.empty_vqip()
        self.storage_ = self.copy_vqip(self.storage)

        self.storage = self.make_decay(self.storage)

    def decay_ds(self):
        """Track storage and amount decayed.

        Returns:
            ds (dict): A VQIP of change in storage and total decayed
        """
        ds = self.ds_vqip(self.storage, self.storage_)
        ds = self.sum_vqip(ds, self.total_decayed)
        return ds


class QueueTank(Tank):
    """"""

    def __init__(self, number_of_timesteps=0, **kwargs):
        """A tank with an internal queue arc, whose queue must be completed before
        storage is available for use. The storage that has completed the queue is under
        the 'active_storage' property.

        Args:
            number_of_timesteps (int, optional): Built in delay for the internal
                queue - it is always added to the queue time, although delay can be
                provided with pushes only. Defaults to 0.
        """
        # Set parameters
        self.number_of_timesteps = number_of_timesteps

        super().__init__(**kwargs)
        self.end_timestep = self._end_timestep
        self.active_storage = self.copy_vqip(self.storage)

        # TODO enable queue to be initialised not empty
        self.out_arcs = {}
        self.in_arcs = {}
        # Create internal queue arc
        self.internal_arc = AltQueueArc(
            in_port=self, out_port=self, number_of_timesteps=self.number_of_timesteps
        )
        # TODO should mass balance call internal arc (is this arc called in arc mass
        #   balance?)

    def apply_overrides(self, overrides: Dict[str, Any] = {}):
        """Apply overrides to the queuetank.

        Enables a user to override any of the following parameters:
        number_of_timesteps.

        Args:
            overrides (dict, optional): Dictionary of overrides. Defaults to {}.
        """
        self.number_of_timesteps = overrides.pop(
            "number_of_timesteps", self.number_of_timesteps
        )
        self.internal_arc.number_of_timesteps = self.number_of_timesteps
        super().apply_overrides(overrides)

    def get_avail(self):
        """Return the active_storage of the tank.

        Returns:
            (dict): VQIP of active_storage
        """
        return self.copy_vqip(self.active_storage)

    def push_storage(self, vqip, time=0, force=False):
        """Push storage into QueueTank, applying travel time, unless forced.

        Args:
            vqip (dict): A VQIP of the amount to push
            time (int, optional): Number of timesteps to spend in queue, in addition
                to number_of_timesteps property of internal_arc. Defaults to 0.
            force (bool, optional): Force property that will ignore tank capacity
                and ignore travel time. Defaults to False.

        Returns:
            reply (dict): A VQIP of water that could not be received by the tank
        """
        if force:
            # Directly add request to storage, skipping queue
            self.storage = self.sum_vqip(self.storage, vqip)
            self.active_storage = self.sum_vqip(self.active_storage, vqip)
            return self.empty_vqip()

        # Push to QueueTank
        reply = self.internal_arc.send_push_request(vqip, force=force, time=time)
        # Update storage
        # TODO storage won't be accurately tracking temperature..
        self.storage = self.sum_vqip(
            self.storage, self.v_change_vqip(vqip, vqip["volume"] - reply["volume"])
        )
        return reply

    def pull_storage(self, vqip):
        """Pull storage from the QueueTank, only water in active_storage is available.
        Returning water pulled and updating tank states. Pollutants are removed from
        tank in proportion to 'volume' in vqip (pollutant values in vqip are ignored).

        Args:
            vqip (dict): VQIP amount to pull, only 'volume' property is used

        Returns:
            reply (dict): VQIP amount that was pulled
        """
        # Adjust based on available volume
        reply = min(vqip["volume"], self.active_storage["volume"])

        # Update reply to vqip
        reply = self.v_change_vqip(self.active_storage, reply)

        # Extract from active_storage
        self.active_storage = self.extract_vqip(self.active_storage, reply)

        # Extract from storage
        self.storage = self.extract_vqip(self.storage, reply)

        return reply

    def pull_storage_exact(self, vqip):
        """Pull storage from the QueueTank, only water in active_storage is available.
        Pollutants are removed from tank in according to their values in vqip.

        Args:
            vqip (dict): A VQIP amount to pull

        Returns:
            reply (dict): A VQIP amount successfully pulled
        """
        # Adjust based on available
        reply = self.copy_vqip(vqip)
        for pol in ["volume"] + constants.ADDITIVE_POLLUTANTS:
            reply[pol] = min(reply[pol], self.active_storage[pol])

        # Pull from QueueTank
        self.active_storage = self.extract_vqip(self.active_storage, reply)

        # Extract from storage
        self.storage = self.extract_vqip(self.storage, reply)
        return reply

    def push_check(self, vqip=None, tag="default"):
        """Wrapper for get_excess but applies comparison to volume in VQIP.
        Needed to enable use of internal_arc, which assumes it is connecting nodes .
        rather than tanks.
        NOTE: this is intended only for use with the internal_arc. Pushing to
        QueueTanks should use 'push_storage'.

        Args:
            vqip (dict, optional): VQIP amount to push. Defaults to None.
            tag (str, optional): Tag, see Node, don't think it should actually be
                used for a QueueTank since there are no handlers. Defaults to
                'default'.

        Returns:
            excess (dict): a VQIP amount of excess capacity
        """
        # TODO does behaviour for volume = None need to be defined?
        excess = self.get_excess()
        if vqip is not None:
            excess["volume"] = min(vqip["volume"], excess["volume"])
        return excess

    def push_set(self, vqip, tag="default"):
        """Behaves differently from normal push setting, it assumes sufficient tank
        capacity and receives VQIPs that have reached the END of the internal_arc.
        NOTE: this is intended only for use with the internal_arc. Pushing to
        QueueTanks should use 'push_storage'.

        Args:
            vqip (dict): VQIP amount to push
            tag (str, optional): Tag, see Node, don't think it should actually be
                used for a QueueTank since there are no handlers. Defaults to
                'default'.

        Returns:
            (dict): Returns empty VQIP, indicating all water received (since it
                assumes capacity was checked before entering the internal arc)
        """
        # Update active_storage (since it has reached the end of the internal_arc)
        self.active_storage = self.sum_vqip(self.active_storage, vqip)

        return self.empty_vqip()

    def _end_timestep(self):
        """Wrapper for end_timestep that also ends the timestep in the internal_arc."""
        self.internal_arc.end_timestep()
        self.internal_arc.update_queue()
        self.storage_ = self.copy_vqip(self.storage)

    def reinit(self):
        """Zeros storages and arc."""
        self.internal_arc.reinit()
        self.storage = self.empty_vqip()
        self.storage_ = self.empty_vqip()
        self.active_storage = self.empty_vqip()


class DecayQueueTank(QueueTank):
    """"""

    def __init__(self, decays={}, parent=None, number_of_timesteps=1, **kwargs):
        """Adds a DecayAltArc in QueueTank to enable decay to occur within the
        internal_arc queue.

        Args:
            decays (dict): A dict of dicts containing a key for each pollutant and,
                within that, a key for each parameter (a constant and exponent)
            parent (object): An object that can be used to read temperature data from
            number_of_timesteps (int, optional): Built in delay for the internal
                queue - it is always added to the queue time, although delay can be
                provided with pushes only. Defaults to 0.
        """
        # Initialise QueueTank
        super().__init__(number_of_timesteps=number_of_timesteps, **kwargs)
        # Replace internal_arc with a DecayArcAlt
        self.internal_arc = DecayArcAlt(
            in_port=self,
            out_port=self,
            number_of_timesteps=number_of_timesteps,
            parent=parent,
            decays=decays,
        )

        self.end_timestep = self._end_timestep

    def apply_overrides(self, overrides: Dict[str, Any] = {}):
        """Apply overrides to the decayqueuetank.

        Enables a user to override any of the following parameters:
        number_of_timesteps, decays.

        Args:
            overrides (dict, optional): Dictionary of overrides. Defaults to {}.
        """
        self.number_of_timesteps = overrides.pop(
            "number_of_timesteps", self.number_of_timesteps
        )
        self.internal_arc.number_of_timesteps = self.number_of_timesteps
        self.internal_arc.decays.update(overrides.pop("decays", {}))
        super().apply_overrides(overrides)

    def _end_timestep(self):
        """End timestep wrapper that removes decayed pollutants and calls internal
        arc."""
        # TODO Should the active storage decay if decays are given (probably.. though
        #   that sounds like a nightmare)?
        self.storage = self.extract_vqip(self.storage, self.internal_arc.total_decayed)
        self.storage_ = self.copy_vqip(self.storage)
        self.internal_arc.end_timestep()



# %% [markdown]
# # Customise a RiverReservoir node (.py)
# Note - this script can also be opened in interactive Python if you wanted to
# play around. On the GitHub it is in [docs/demo/scripts](https://github.com/barneydobson/wsi/blob/main/docs/demo/scripts/customise_riverreservoir.py)
#
# 1. [Introduction](#introduction)
#
# 2. [Climate inputs](#climate-inputs)
#
# 3. [Environmental flows](#environmental-flows)
#
# 4. [Infiltration](#infiltration)
#
# 5. [What next](#what-next)
#

# %% [markdown]
# ## Introduction
# In this tutorial we will demonstrate how to customise a
# [`RiverReservoir`](https://barneydobson.github.io/wsi/reference-storage/#wsimod.nodes.storage.RiverReservoir)
# node. This is a run-of-river reservoir, but only a minimal implementation
# is provided, which does not capture many of the processes that WSIMOD users
# are likely to require.
#
# Rather than customise the `RiverReservoir` node, we would prefer to equip
# users with the ability to customise WSIMOD behaviour to accommodate the
# almost infinite range of possibilities of how would you like a node to behave.
#
# Thus in this guide, we will customise a `RiverReservoir` node to include the following
# processes:
#  - Precipitation/evapotranspiration from the reservoir surface. This is an example of
#  assigning input data to a node.
#  - Ensure that environmental flows are satisfied. This is an example of calling an
#  additional function during the orchestration of a node.
#  - Infiltration to attached groundwater nodes. This is an example of adding new behaviour
#  to a node

# %%
# Import packages
from pprint import pprint as print

from wsimod.arcs.arcs import Arc
from wsimod.core import constants
from wsimod.nodes.storage import Groundwater, RiverReservoir
from wsimod.nodes.waste import Waste
from wsimod.orchestration.model import Model, to_datetime

# Set simple pollutants
constants.set_simple_pollutants()
# %% [markdown]
# We will create a simple model that includes a reservoir, a groundwater, and an outlet.
# %%
# Create nodes
my_reservoir = RiverReservoir(
    name="my_reservoir", capacity=10, area=5, initial_storage=6, environmental_flow=1
)

my_groundwater = Groundwater(
    name="my_groundwater", capacity=100, area=10, initial_storage=50
)

my_outlet = Waste(name="my_outlet")

# %% [markdown]
# We provide arcs for reservoir outflow (i.e., the environmental flow), the infiltration which we
# will later include, and a baseflow of groundwater.

# %%
# Create arcs
infiltration_arc = Arc(
    name="infiltration", in_port=my_reservoir, out_port=my_groundwater
)

environmental_arc = Arc(name="environmental", in_port=my_reservoir, out_port=my_outlet)

baseflow_arc = Arc(name="baseflow", in_port=my_groundwater, out_port=my_outlet)

# %%
# Create and populate model
my_model = Model()

my_model.add_instantiated_nodes([my_reservoir, my_groundwater, my_outlet])

my_model.add_instantiated_arcs([infiltration_arc, environmental_arc, baseflow_arc])

# Assign a date to run model for
date = to_datetime("2000-01-01")

# %%
# Run model
results = my_model.run(dates=[date])

# Inspect flows
flow = results[0]
print(flow)

stores = results[1]
print(stores)

# %% [markdown]
# We can see here that the only process that has happened is some groundwater baseflow.
# Neither the climate forcing, nor environmental flow, nor infiltration occur by default,
# despite that we have included arcs to connect them. This is because in the case of
# climate flow and infiltration, the behaviour is not yet defined. While, for environmental
# flows, we can see that a [`satisfy_environmental`](https://barneydobson.github.io/wsi/reference-storage/#wsimod.nodes.storage.RiverReservoir.satisfy_environmental)
# function is provided in the API - however it is not called in the default model orchestration.
#
# Thus, we will customise our node to address all of these points.
# %% [markdown]
# ## Climate inputs

# %% [markdown]
# ### Define inputs
# We need to define some precipitation/evapotranspiration climate data to
# force the reservoir with, and then store that information.

# %%
# Create climate data
climate_data = {("precipitation", date): 0.01, ("et0", date): 0.005}

# Give reservoir node access to it
my_reservoir.data_input_dict = climate_data

# %% [markdown]
# ### Mass balance
# The first thing we should always consider is how our change will impact mass
# balance checking, since there is nothing more annoying than our model throwing
# a bunch of mass balance errors.
#
# To do this we will create two state variables `net_evaporation` and
# `net_precipitation` that belong to the `my_reservoir` node. Every object in
# WSIMOD has three lists related to mass balance checking: `mass_balance_in`,
# `mass_balance_out`, and `mass_balance_ds`. You can read more about it in the [API](https://barneydobson.github.io/wsi/reference-core/#wsimod.core.core.WSIObj.mass_balance)
# but in short, any changes we make need to be appended as functions to these
# lists.
#
# Thus, we store these state variables in a `lambda` call in the lists. Note that
# we have given the `self = my_reservoir` argument in the `lamdba` call to ensure
# that the function gets stored in the node's memory, rather than hard coded to
# this point in the script.
# %%
# Add mass balance terms
my_reservoir.net_evaporation = my_reservoir.empty_vqip()
my_reservoir.mass_balance_out.append(lambda self=my_reservoir: self.net_evaporation)

my_reservoir.net_precipitation = my_reservoir.empty_vqip()
my_reservoir.mass_balance_in.append(lambda self=my_reservoir: self.net_precipitation)


# %% [markdown]
# ### Add new behaviour
# Now we need to add the processes by which evaporation/precipitation happens to
# the surface of the reservoir. Our first task is to figure out how this behaviour might be called during orchestration. If we inspect the source code
# in the [`Model`](https://barneydobson.github.io/wsi/reference-model/#wsimod.orchestration.model.Model.run)
# object, we will see what functions are called during orchestration (i.e., the
# `run` function of the `Model`). We can see that the only function called for
# `Reservoir` objects is the `make_abstractions` function. Thus, the simplest way
# for us to add new behaviour to a `Reservoir` is to 'decorate' this function.
#
# Note that this is only the right solution if we are happy that these functions
# take place at this point of orchestration. For example, we might be happy for
# the climate processes to occur at any point in the timestep, but if we view the
# order of operations, we might be less sure that infiltration occurs after
# groundwater has been distributed. For the purpose of this demo though, we will
# assume that we are happy for all of our new reservoir behaviour to take place
# at the same time within a timestep.
#
# As mentioned, we must decorate our `make_abstractions` function. To do this we
# provide a wrapper for the new behaviour that also calls the original function.
# Then we overwrite the memory address of the original function in the object to
# perform this new wrapped function. Decorators can be a bit to get your head around, you can read more about them at: https://www.thecodeship.com/patterns/guide-to-python-function-decorators/.
#
# The equations that we need to add for this are reasonably self-explanatory, and
# take place inside the `reservoir_functions_wrapper`.
# %%
# Decorate reservoir to include climate processes
def wrapper(node, func):
    """

    Args:
        node:
        func:

    Returns:

    """

    def reservoir_functions_wrapper():
        """

        Returns:

        """
        # Initialise mass balance VQIPs
        vqip_out = node.empty_vqip()
        vqip_in = node.empty_vqip()

        # Calculate net change
        net_in = node.get_data_input("precipitation") - node.get_data_input("et0")
        net_in *= node.tank.area

        if net_in > 0:
            # Add precipitation
            vqip_in = node.v_change_vqip(node.empty_vqip(), net_in)
            _ = node.tank.push_storage(vqip_in, force=True)

        else:
            # Remove evaporation
            evap = node.tank.evaporate(-net_in)
            vqip_out = node.v_change_vqip(vqip_out, evap)

        # Store in mass balance states
        my_reservoir.net_evaporation = vqip_out
        my_reservoir.net_precipitation = vqip_in

        # Call whatever else was going happen
        return func()

    return reservoir_functions_wrapper


# Run decorator
my_reservoir.make_abstractions = wrapper(my_reservoir, my_reservoir.make_abstractions)

# %% [markdown]
# OK fantastic, there's no new flows along the arcs, though the baseflow has reduced a bit because it is based on the total groundwater storage which has decreased in our previous example.
#
# Most importantly we see that the reservoir volume has increased by 0.025m3, thus our new functionality seems to work. Next we will include environmental
# flows.

# %%
# Reinspect results
results = my_model.run(dates=[date])

flow = results[0]
print(flow)

stores = results[1]
print(stores)


# %% [markdown]
# ## Environmental flows
# As mentioned, the `RiverReservoir` object already has a parameter
# (`environmental_flow`) and a function (`satisfy_environmental`) to carry out
# releases for environmental flows. So all we need to do is redecorate our
# `make_abstractions` function to call this new function.
# %%
# Decorate reservoir to include environmental flows
def wrapper(node, func):
    """

    Args:
        node:
        func:

    Returns:

    """

    def reservoir_functions_wrapper():
        """

        Returns:

        """
        node.satisfy_environmental()
        return func()

    return reservoir_functions_wrapper


my_reservoir.make_abstractions = wrapper(my_reservoir, my_reservoir.make_abstractions)

# %% [markdown]
# Notice how we were able to just pass the decorated version of the
# `make_abstractions` function, since that version of the function is still
# containing all of the new climate behaviour. In practice, this isn't very transparent and we would make one wrapper that includes all of the new
# functionality that we want to add, but it is fine for the demo.
#
# We now see that the `environmental_flow` of 1 has been satisfied along the arc from the reservoir to the outlet. We also see that the reservoir storage has decreased by this amount but increased by the precipitation amount again, which means our climate functions are still working fine.
#
# We also see an incredibly small amount of flow along the infiltration arc, this is because of how the `push_distributed` function called by `satisfy_environmental` works, which is behaviour for another tutorial, but can be considered negligible for this demo.
# %%
# Reinspect results
results = my_model.run(dates=[date])

flow = results[0]

print(flow)

stores = results[1]
print(stores)

# %% [markdown]
# ## Infiltration
# Our final task is to implement infiltration. We are cool hands at decorating
# now so should feel confident that we simply need to add some new functionality
# within a wrapper and decorate the `make_abstractions` function.
# %%
# Decorate reservoir to include infiltration
my_reservoir.infiltration_rate = 0.002  # [metres per timestep]


def wrapper(node, func):
    """

    Args:
        node:
        func:

    Returns:

    """

    def reservoir_functions_wrapper():
        """

        Returns:

        """
        # Calculate infiltration amount
        infiltration_amount = node.infiltration_rate * node.tank.area

        # Extract from reservoir tank
        infiltration_vqip = node.tank.pull_storage({"volume": infiltration_amount})

        # Distribute to any connected groundwater nodes
        reply = node.push_distributed(infiltration_vqip, of_type="Groundwater")

        # If any was not successfully received, store this back in the reservoir
        node.tank.push_storage(reply, force=True)
        return func()

    return reservoir_functions_wrapper


my_reservoir.make_abstractions = wrapper(my_reservoir, my_reservoir.make_abstractions)
# %% [markdown]
# Sure enough, our infiltration works!
# %%
# Reinspect results
results = my_model.run(dates=[date])

flow = results[0]

print(flow)

stores = results[1]
print(stores)

# %% [markdown]
# ## What next?
# If you found this a little complicated, don't worry, we were doing some quite
# complicated things to our nodes. In general, it is easier to customise the
# interactions between nodes than the nodes themselves. You can find out more
# about this in our
# [`customise_interactions`](https://barneydobson.github.io/wsi/demo/scripts/customise_interactions/)
# guide!



# -*- coding: utf-8 -*-
"""Created on Mon Jul  4 16:01:48 2022.

@author: bdobson
"""
import csv
import gzip
import inspect
import os
import sys
from datetime import datetime
from math import log10

import dill as pickle
import yaml
from tqdm import tqdm

from wsimod.arcs import arcs as arcs_mod
from wsimod.core import constants
from wsimod.core.core import WSIObj
from wsimod.nodes.land import ImperviousSurface
from wsimod.nodes.nodes import NODES_REGISTRY
from wsimod.nodes.tanks import QueueTank, ResidenceTank, Tank

os.environ["USE_PYGEOS"] = "0"


class to_datetime:
    """"""

    # TODO document and make better
    def __init__(self, date_string):
        """Simple datetime wrapper that has key properties used in WSIMOD components.

        Args:
            date_string (str): A string containing the date, expected in
                format %Y-%m-%d or %Y-%m.
        """
        self._date = self._parse_date(date_string)

    def __str__(self):
        return self._date.strftime("%Y-%m-%d")

    def __repr__(self):
        return self._date.strftime("%Y-%m-%d")

    @property
    def dayofyear(self):
        """

        Returns:

        """
        return self._date.timetuple().tm_yday

    @property
    def day(self):
        """

        Returns:

        """
        return self._date.day

    @property
    def year(self):
        """

        Returns:

        """
        return self._date.year

    @property
    def month(self):
        """

        Returns:

        """
        return self._date.month

    def to_period(self, args="M"):
        """

        Args:
            args:

        Returns:

        """
        return to_datetime(f"{self._date.year}-{str(self._date.month).zfill(2)}")

    def is_leap_year(self):
        """

        Returns:

        """
        year = self._date.year
        return year % 4 == 0 and (year % 100 != 0 or year % 400 == 0)

    def _parse_date(self, date_string, date_format="%Y-%m-%d %H:%M:%S"):
        try:
            return datetime.strptime(date_string, date_format)
        except ValueError:
            try:
                return datetime.strptime(date_string, "%Y-%m-%d")
            except ValueError:
                try:
                    # Check if valid 'YYYY-MM' format
                    if len(date_string.split("-")[0]) == 4:
                        int(date_string.split("-")[0])
                    if len(date_string.split("-")[1]) == 2:
                        int(date_string.split("-")[1])
                    return date_string
                except ValueError:
                    raise ValueError

    def __eq__(self, other):
        if isinstance(other, to_datetime):
            return self._date == other._date
        return False

    def __hash__(self):
        return hash(self._date)


class Model(WSIObj):
    """"""

    def __init__(self):
        """Object to contain nodes and arcs that provides a default orchestration.

        Returns:
            Model: An empty model object
        """
        super().__init__()
        self.arcs = {}
        # self.arcs_type = {} #not sure that this would be necessary
        self.nodes = {}
        self.nodes_type = {}
        self.extensions = []
        self.river_discharge_order = []

        # Default orchestration
        self.orchestration = [
            {"FWTW": "treat_water"},
            {"Demand": "create_demand"},
            {"Land": "run"},
            {"Groundwater": "infiltrate"},
            {"Sewer": "make_discharge"},
            {"Foul": "make_discharge"},
            {"WWTW": "calculate_discharge"},
            {"Groundwater": "distribute"},
            {"River": "calculate_discharge"},
            {"Reservoir": "make_abstractions"},
            {"Land": "apply_irrigation"},
            {"WWTW": "make_discharge"},
            {"Catchment": "route"},
        ]

    def get_init_args(self, cls):
        """Get the arguments of the __init__ method for a class and its superclasses."""
        init_args = []
        for c in cls.__mro__:
            # Get the arguments of the __init__ method
            args = inspect.getfullargspec(c.__init__).args[1:]
            init_args.extend(args)
        return init_args

    def load(self, address, config_name="config.yml", overrides={}):
        """

        Args:
            address:
            config_name:
            overrides:
        """
        from ..extensions import apply_patches

        with open(os.path.join(address, config_name), "r") as file:
            data: dict = yaml.safe_load(file)

        for key, item in overrides.items():
            data[key] = item

        constants.POLLUTANTS = data.get("pollutants", constants.POLLUTANTS)
        constants.ADDITIVE_POLLUTANTS = data.get(
            "additive_pollutants", constants.ADDITIVE_POLLUTANTS
        )
        constants.NON_ADDITIVE_POLLUTANTS = data.get(
            "non_additive_pollutants", constants.NON_ADDITIVE_POLLUTANTS
        )
        constants.FLOAT_ACCURACY = float(
            data.get("float_accuracy", constants.FLOAT_ACCURACY)
        )
        self.__dict__.update(Model().__dict__)

        """
        FLAG:
            E.G. ADDITION FOR NEW ORCHESTRATION
        """
        load_extension_files(data.get("extensions", []))
        self.extensions = data.get("extensions", [])

        if "orchestration" in data.keys():
            # Update orchestration
            self.orchestration = data["orchestration"]

        if "nodes" not in data.keys():
            raise ValueError("No nodes found in the config")

        nodes = data["nodes"]

        for name, node in nodes.items():
            if "filename" in node.keys():
                node["data_input_dict"] = read_csv(
                    os.path.join(address, node["filename"])
                )
                del node["filename"]
            if "surfaces" in node.keys():
                for key, surface in node["surfaces"].items():
                    if "filename" in surface.keys():
                        node["surfaces"][key]["data_input_dict"] = read_csv(
                            os.path.join(address, surface["filename"])
                        )
                        del surface["filename"]
                node["surfaces"] = list(node["surfaces"].values())
        arcs = data.get("arcs", {})
        self.add_nodes(list(nodes.values()))
        self.add_arcs(list(arcs.values()))

        self.add_overrides(data.get("overrides", {}))

        if "dates" in data.keys():
            self.dates = [to_datetime(x) for x in data["dates"]]

        apply_patches(self)

    def save(self, address, config_name="config.yml", compress=False):
        """Save the model object to a yaml file and input data to csv.gz format in the
        directory specified.

        Args:
            address (str): Path to a directory
            config_name (str, optional): Name of yaml model file.
                Defaults to 'model.yml'
        """
        if not os.path.exists(address):
            os.mkdir(address)
        nodes = {}

        if compress:
            file_type = "csv.gz"
        else:
            file_type = "csv"
        for node in self.nodes.values():
            init_args = self.get_init_args(node.__class__)
            special_args = set(["surfaces", "parent", "data_input_dict"])

            node_props = {
                x: getattr(node, x) for x in set(init_args).difference(special_args)
            }
            node_props["type_"] = node.__class__.__name__
            node_props["node_type_override"] = (
                repr(node.__class__).split(".")[-1].replace("'>", "")
            )

            if "surfaces" in init_args:
                surfaces = {}
                for surface in node.surfaces:
                    surface_args = self.get_init_args(surface.__class__)
                    surface_props = {
                        x: getattr(surface, x)
                        for x in set(surface_args).difference(special_args)
                    }
                    surface_props["type_"] = surface.__class__.__name__

                    # Exceptions...
                    # TODO I need a better way to do this
                    del surface_props["capacity"]
                    if set(["rooting_depth", "pore_depth"]).intersection(surface_args):
                        del surface_props["depth"]
                    if "data_input_dict" in surface_args:
                        if surface.data_input_dict:
                            filename = (
                                "{0}-{1}-inputs.{2}".format(
                                    node.name, surface.surface, file_type
                                )
                                .replace("(", "_")
                                .replace(")", "_")
                                .replace("/", "_")
                                .replace(" ", "_")
                            )
                            write_csv(
                                surface.data_input_dict,
                                {"node": node.name, "surface": surface.surface},
                                os.path.join(address, filename),
                                compress=compress,
                            )
                            surface_props["filename"] = filename
                    surfaces[surface_props["surface"]] = surface_props
                node_props["surfaces"] = surfaces

            if "data_input_dict" in init_args:
                if node.data_input_dict:
                    filename = "{0}-inputs.{1}".format(node.name, file_type)
                    write_csv(
                        node.data_input_dict,
                        {"node": node.name},
                        os.path.join(address, filename),
                        compress=compress,
                    )
                    node_props["filename"] = filename

            nodes[node.name] = node_props

        arcs = {}
        for arc in self.arcs.values():
            init_args = self.get_init_args(arc.__class__)
            special_args = set(["in_port", "out_port"])
            arc_props = {
                x: getattr(arc, x) for x in set(init_args).difference(special_args)
            }
            arc_props["type_"] = arc.__class__.__name__
            arc_props["in_port"] = arc.in_port.name
            arc_props["out_port"] = arc.out_port.name
            arcs[arc.name] = arc_props

        data = {
            "nodes": nodes,
            "arcs": arcs,
            "orchestration": self.orchestration,
            "pollutants": constants.POLLUTANTS,
            "additive_pollutants": constants.ADDITIVE_POLLUTANTS,
            "non_additive_pollutants": constants.NON_ADDITIVE_POLLUTANTS,
            "float_accuracy": constants.FLOAT_ACCURACY,
            "extensions": self.extensions,
            "river_discharge_order": self.river_discharge_order,
        }
        if hasattr(self, "dates"):
            data["dates"] = [str(x) for x in self.dates]

        def coerce_value(value):
            """

            Args:
                value:

            Returns:

            """
            conversion_options = {
                "__float__": float,
                "__iter__": list,
                "__int__": int,
                "__str__": str,
                "__bool__": bool,
            }
            converted = False
            for property, func in conversion_options.items():
                if hasattr(value, property):
                    try:
                        yaml.safe_dump(func(value))
                        value = func(value)
                        converted = True
                        break
                    except Exception:
                        raise ValueError(f"Cannot dump: {value} of type {type(value)}")
            if not converted:
                raise ValueError(f"Cannot dump: {value} of type {type(value)}")

            return value

        def check_and_coerce_dict(data_dict):
            """

            Args:
                data_dict:
            """
            for key, value in data_dict.items():
                if isinstance(value, dict):
                    check_and_coerce_dict(value)
                else:
                    try:
                        yaml.safe_dump(value)
                    except yaml.representer.RepresenterError:
                        if hasattr(value, "__iter__"):
                            for idx, val in enumerate(value):
                                if isinstance(val, dict):
                                    check_and_coerce_dict(val)
                                else:
                                    value[idx] = coerce_value(val)
                        data_dict[key] = coerce_value(value)

        check_and_coerce_dict(data)

        write_yaml(address, config_name, data)

    def load_pickle(self, fid):
        """Load model object to a pickle file, including the model states.

        Args:
            fid (str): File address to load the pickled model from

        Returns:
            model (obj): loaded model

        Example:
            >>> # Load and run your model
            >>> my_model.load(model_dir,config_name = 'config.yml')
            >>> _ = my_model.run()
            >>>
            >>> # Save it including its different states
            >>> my_model.save_pickle('model_at_end_of_run.pkl')
            >>>
            >>> # Load it at another time to resume the model from the end
            >>> # of the previous run
            >>> new_model = Model()
            >>> new_model = new_model.load_pickle('model_at_end_of_run.pkl')
        """
        file = open(fid, "rb")
        return pickle.load(file)

    def save_pickle(self, fid):
        """Save model object to a pickle file, including saving the model states.

        Args:
            fid (str): File address to save the pickled model to

        Returns:
            message (str): Exit message of pickle dump
        """
        file = open(fid, "wb")
        pickle.dump(self, file)
        return file.close()

    def add_nodes(self, nodelist):
        """Add nodes to the model object from a list of dicts, where each dict contains
        all of the parameters for a node. Intended to be called before add_arcs.

        Args:
            nodelist (list): List of dicts, where a dict is a node
        """

        for data in nodelist:
            name = data["name"]
            type_ = data["type_"]
            if "node_type_override" in data.keys():
                node_type = data["node_type_override"]
                del data["node_type_override"]
            else:
                node_type = type_
            if "foul" in name:
                # Absolute hack to enable foul sewers to be treated separate from storm
                type_ = "Foul"
            if "geometry" in data.keys():
                del data["geometry"]
            del data["type_"]

            if node_type not in NODES_REGISTRY.keys():
                raise ValueError(f"Node type {node_type} not recognised")

            if type_ not in self.nodes_type.keys():
                self.nodes_type[type_] = {}

            self.nodes_type[type_][name] = NODES_REGISTRY[node_type](**dict(data))
            self.nodes[name] = self.nodes_type[type_][name]
            self.nodelist = [x for x in self.nodes.values()]

    def add_instantiated_nodes(self, nodelist):
        """Add nodes to the model object from a list of objects, where each object is an
        already instantiated node object. Intended to be called before add_arcs.

        Args:
            nodelist (list): list of objects that are nodes
        """
        self.nodelist = nodelist
        self.nodes = {x.name: x for x in nodelist}
        for x in nodelist:
            type_ = x.__class__.__name__
            if type_ not in self.nodes_type.keys():
                self.nodes_type[type_] = {}
            self.nodes_type[type_][x.name] = x

    def add_arcs(self, arclist):
        """Add nodes to the model object from a list of dicts, where each dict contains
        all of the parameters for an arc.

        Args:
            arclist (list): list of dicts, where a dict is an arc
        """
        river_arcs = {}
        for arc in arclist:
            name = arc["name"]
            type_ = arc["type_"]
            del arc["type_"]
            arc["in_port"] = self.nodes[arc["in_port"]]
            arc["out_port"] = self.nodes[arc["out_port"]]
            self.arcs[name] = getattr(arcs_mod, type_)(**dict(arc))

            if arc["in_port"].__class__.__name__ in [
                "River",
                "Node",
                "Waste",
                "Reservoir",
            ]:
                if arc["out_port"].__class__.__name__ in [
                    "River",
                    "Node",
                    "Waste",
                    "Reservoir",
                ]:
                    river_arcs[name] = self.arcs[name]

        self.river_discharge_order = []
        if not any(river_arcs):
            return
        upstreamness = (
            {x: 0 for x in self.nodes_type["Waste"].keys()}
            if "Waste" in self.nodes_type
            else {}
        )
        upstreamness = self.assign_upstream(river_arcs, upstreamness)

        if "River" in self.nodes_type:
            for node in sorted(
                upstreamness.items(), key=lambda item: item[1], reverse=True
            ):
                if node[0] in self.nodes_type["River"]:
                    self.river_discharge_order.append(node[0])

    def add_instantiated_arcs(self, arclist):
        """Add arcs to the model object from a list of objects, where each object is an
        already instantiated arc object.

        Args:
            arclist (list): list of objects that are arcs.
        """
        self.arclist = arclist
        self.arcs = {x.name: x for x in arclist}
        river_arcs = {}
        for arc in arclist:
            if arc.in_port.__class__.__name__ in [
                "River",
                "Node",
                "Waste",
                "Reservoir",
            ]:
                if arc.out_port.__class__.__name__ in [
                    "River",
                    "Node",
                    "Waste",
                    "Reservoir",
                ]:
                    river_arcs[arc.name] = arc
        if not any(river_arcs):
            return
        upstreamness = (
            {x: 0 for x in self.nodes_type["Waste"].keys()}
            if "Waste" in self.nodes_type
            else {}
        )
        upstreamness = {x: 0 for x in self.nodes_type["Waste"].keys()}

        upstreamness = self.assign_upstream(river_arcs, upstreamness)

        self.river_discharge_order = []
        if "River" in self.nodes_type:
            for node in sorted(
                upstreamness.items(), key=lambda item: item[1], reverse=True
            ):
                if node[0] in self.nodes_type["River"]:
                    self.river_discharge_order.append(node[0])

    def assign_upstream(self, arcs, upstreamness):
        """Recursive function to trace upstream up arcs to determine which are the most
        upstream.

        Args:
            arcs (list): list of dicts where dicts are arcs
            upstreamness (dict): dictionary contain nodes in
                arcs as keys and a number representing upstreamness
                (higher numbers = more upstream)

        Returns:
            upstreamness (dict): final version of upstreamness
        """
        upstreamness_ = upstreamness.copy()
        in_nodes = [
            x.in_port.name
            for x in arcs.values()
            if x.out_port.name in upstreamness.keys()
        ]
        ind = max(list(upstreamness_.values())) + 1
        in_nodes = list(set(in_nodes).difference(upstreamness.keys()))
        for node in in_nodes:
            upstreamness[node] = ind
        if upstreamness == upstreamness_:
            return upstreamness
        else:
            upstreamness = self.assign_upstream(arcs, upstreamness)
            return upstreamness

    def add_overrides(self, config: dict):
        """Apply overrides to nodes and arcs in the model object.

        Args:
            config (dict): dictionary of overrides to apply to the model object.
        """
        for node in config.get("nodes", {}).values():
            type_ = node.pop("type_")
            name = node.pop("name")

            if type_ not in self.nodes_type.keys():
                raise ValueError(f"Node type {type_} not recognised")

            if name not in self.nodes_type[type_].keys():
                raise ValueError(f"Node {name} not recognised")

            self.nodes_type[type_][name].apply_overrides(node)

        for arc in config.get("arcs", {}).values():
            name = arc.pop("name")
            type_ = arc.pop("type_")

            if name not in self.arcs.keys():
                raise ValueError(f"Arc {name} not recognised")

            self.arcs[name].apply_overrides(arc)

    def debug_node_mb(self):
        """Simple function that iterates over nodes calling their mass balance
        function."""
        for node in self.nodelist:
            _ = node.node_mass_balance()

    def default_settings(self):
        """Incomplete function that enables easy specification of results storage.

        Returns:
            (dict): default settings
        """
        return {
            "arcs": {"flows": True, "pollutants": True},
            "tanks": {"storages": True, "pollutants": True},
            "mass_balance": False,
        }

    def change_runoff_coefficient(self, relative_change, nodes=None):
        """Clunky way to change the runoff coefficient of a land node.

        Args:
            relative_change (float): amount that the impervious area in the land
                node is multiplied by (grass area is changed in compensation)
            nodes (list, optional): list of land nodes to change the parameters of.
                Defaults to None, which applies the change to all land nodes.
        """
        # Multiplies impervious area by relative change and adjusts grassland
        # accordingly
        if nodes is None:
            nodes = self.nodes_type["Land"].values()

        if isinstance(relative_change, float):
            relative_change = {x: relative_change for x in nodes}

        for node in nodes:
            surface_dict = {x.surface: x for x in node.surfaces}
            if "Impervious" in surface_dict.keys():
                impervious_area = surface_dict["Impervious"].area
                grass_area = surface_dict["Grass"].area

                new_impervious_area = impervious_area * relative_change[node]
                new_grass_area = grass_area + (impervious_area - new_impervious_area)
                if new_grass_area < 0:
                    print("not enough grass")
                    break
                surface_dict["Impervious"].area = new_impervious_area
                surface_dict["Impervious"].capacity *= relative_change[node]

                surface_dict["Grass"].area = new_grass_area
                surface_dict["Grass"].capacity *= new_grass_area / grass_area
                for pol in constants.ADDITIVE_POLLUTANTS + ["volume"]:
                    surface_dict["Grass"].storage[pol] *= new_grass_area / grass_area
                for pool in surface_dict["Grass"].nutrient_pool.pools:
                    for nutrient in pool.storage.keys():
                        pool.storage[nutrient] *= new_grass_area / grass_area

    def run(
        self,
        dates=None,
        settings=None,
        record_arcs=None,
        record_tanks=None,
        record_surfaces=None,
        verbose=True,
        record_all=True,
        objectives=[],
    ):
        """Run the model object with the default orchestration.

        Args:
            dates (list, optional): Dates to simulate. Defaults to None, which
                simulates all dates that the model has data for.
            settings (dict, optional): Dict to specify what results are stored,
                not currently used. Defaults to None.
            record_arcs (list, optional): List of arcs to store result for.
                Defaults to None.
            record_tanks (list, optional): List of nodes with water stores to
                store results for. Defaults to None.
            record_surfaces (list, optional): List of tuples of
                (land node, surface) to store results for. Defaults to None.
            verbose (bool, optional): Prints updates on simulation if true.
                Defaults to True.
            record_all (bool, optional): Specifies to store all results.
                Defaults to True.
            objectives (list, optional): A list of dicts with objectives to
                calculate (see examples). Defaults to [].

        Returns:
            flows: simulated flows in a list of dicts
            tanks: simulated tanks storages in a list of dicts
            objective_results: list of values based on objectives list
            surfaces: simulated surface storages of land nodes in a list of dicts

        Examples:
            # Run a model without storing any results but calculating objectives
            import statistics as stats
            objectives = [{'element_type' : 'flows',
                           'name' : 'my_river',
                           'function' : @ (x, _) stats.mean([y['phosphate'] for y in x])
                           },
                          {'element_type' : 'tanks',
                           'name' : 'my_reservoir',
                           'function' : @ (x, model) sum([y['storage'] < (model.nodes
                           ['my_reservoir'].tank.capacity / 2) for y in x])
                           }]
            _, _, results, _ = my_model.run(record_all = False, objectives = objectives)
        """
        if record_arcs is None:
            record_arcs = []
            if record_all:
                record_arcs = list(self.arcs.keys())
        if record_tanks is None:
            record_tanks = []

        if record_surfaces is None:
            record_surfaces = []

        if settings is None:
            settings = self.default_settings()

        def blockPrint():
            """

            Returns:

            """
            stdout = sys.stdout
            sys.stdout = open(os.devnull, "w")
            return stdout

        def enablePrint(stdout):
            """

            Args:
                stdout:
            """
            sys.stdout = stdout

        if not verbose:
            stdout = blockPrint()
        if dates is None:
            dates = self.dates

        for objective in objectives:
            if objective["element_type"] == "tanks":
                record_tanks.append(objective["name"])
            elif objective["element_type"] == "flows":
                record_arcs.append(objective["name"])
            elif objective["element_type"] == "surfaces":
                record_surfaces.append((objective["name"], objective["surface"]))
            else:
                print("element_type not recorded")

        flows = []
        tanks = []
        surfaces = []
        for date in tqdm(dates, disable=(not verbose)):
            # for date in dates:
            for node in self.nodelist:
                node.t = date
                node.monthyear = date.to_period("M")

            # Iterate over orchestration
            for timestep_item in self.orchestration:
                for node_type, function in timestep_item.items():
                    for node in self.nodes_type.get(node_type, {}).values():
                        getattr(node, function)()

            # river
            for node_name in self.river_discharge_order:
                self.nodes[node_name].distribute()

            # mass balance checking
            # nodes/system
            sys_in = self.empty_vqip()
            sys_out = self.empty_vqip()
            sys_ds = self.empty_vqip()

            # arcs
            for arc in self.arcs.values():
                in_, ds_, out_ = arc.arc_mass_balance()
                for v in constants.ADDITIVE_POLLUTANTS + ["volume"]:
                    sys_in[v] += in_[v]
                    sys_out[v] += out_[v]
                    sys_ds[v] += ds_[v]
            for node in self.nodelist:
                # print(node.name)
                in_, ds_, out_ = node.node_mass_balance()

                # temp = {'name' : node.name,
                #         'time' : date}
                # for lab, dict_ in zip(['in','ds','out'], [in_, ds_, out_]):
                #     for key, value in dict_.items():
                #         temp[(lab, key)] = value
                # node_mb.append(temp)

                for v in constants.ADDITIVE_POLLUTANTS + ["volume"]:
                    sys_in[v] += in_[v]
                    sys_out[v] += out_[v]
                    sys_ds[v] += ds_[v]

            for v in constants.ADDITIVE_POLLUTANTS + ["volume"]:
                # Find the largest value of in_, out_, ds_
                largest = max(sys_in[v], sys_in[v], sys_in[v])

                if largest > constants.FLOAT_ACCURACY:
                    # Convert perform comparison in a magnitude to match the largest
                    # value
                    magnitude = 10 ** int(log10(largest))
                    in_10 = sys_in[v] / magnitude
                    out_10 = sys_in[v] / magnitude
                    ds_10 = sys_in[v] / magnitude
                else:
                    in_10 = sys_in[v]
                    ds_10 = sys_in[v]
                    out_10 = sys_in[v]

                if (in_10 - ds_10 - out_10) > constants.FLOAT_ACCURACY:
                    print(
                        "system mass balance error for "
                        + v
                        + " of "
                        + str(sys_in[v] - sys_ds[v] - sys_out[v])
                    )

            # Store results
            for arc in record_arcs:
                arc = self.arcs[arc]
                flows.append(
                    {"arc": arc.name, "flow": arc.vqip_out["volume"], "time": date}
                )
                for pol in constants.POLLUTANTS:
                    flows[-1][pol] = arc.vqip_out[pol]

            for node in record_tanks:
                node = self.nodes[node]
                tanks.append(
                    {
                        "node": node.name,
                        "storage": node.tank.storage["volume"],
                        "time": date,
                    }
                )

            for node, surface in record_surfaces:
                node = self.nodes[node]
                name = node.name
                surface = node.get_surface(surface)
                if not isinstance(surface, ImperviousSurface):
                    surfaces.append(
                        {
                            "node": name,
                            "surface": surface.surface,
                            "percolation": surface.percolation["volume"],
                            "subsurface_r": surface.subsurface_flow["volume"],
                            "surface_r": surface.infiltration_excess["volume"],
                            "storage": surface.storage["volume"],
                            "evaporation": surface.evaporation["volume"],
                            "precipitation": surface.precipitation["volume"],
                            "tank_recharge": surface.tank_recharge,
                            "capacity": surface.capacity,
                            "time": date,
                            "et0_coef": surface.et0_coefficient,
                            # 'crop_factor' : surface.crop_factor
                        }
                    )
                    for pol in constants.POLLUTANTS:
                        surfaces[-1][pol] = surface.storage[pol]
                else:
                    surfaces.append(
                        {
                            "node": name,
                            "surface": surface.surface,
                            "storage": surface.storage["volume"],
                            "evaporation": surface.evaporation["volume"],
                            "precipitation": surface.precipitation["volume"],
                            "capacity": surface.capacity,
                            "time": date,
                        }
                    )
                    for pol in constants.POLLUTANTS:
                        surfaces[-1][pol] = surface.storage[pol]
            if record_all:
                for node in self.nodes.values():
                    for prop_ in dir(node):
                        prop = node.__getattribute__(prop_)
                        if prop.__class__ in [QueueTank, Tank, ResidenceTank]:
                            tanks.append(
                                {
                                    "node": node.name,
                                    "time": date,
                                    "storage": prop.storage["volume"],
                                    "prop": prop_,
                                }
                            )
                            for pol in constants.POLLUTANTS:
                                tanks[-1][pol] = prop.storage[pol]

                for name, node in self.nodes_type.get("Land", {}).items():
                    for surface in node.surfaces:
                        if not isinstance(surface, ImperviousSurface):
                            surfaces.append(
                                {
                                    "node": name,
                                    "surface": surface.surface,
                                    "percolation": surface.percolation["volume"],
                                    "subsurface_r": surface.subsurface_flow["volume"],
                                    "surface_r": surface.infiltration_excess["volume"],
                                    "storage": surface.storage["volume"],
                                    "evaporation": surface.evaporation["volume"],
                                    "precipitation": surface.precipitation["volume"],
                                    "tank_recharge": surface.tank_recharge,
                                    "capacity": surface.capacity,
                                    "time": date,
                                    "et0_coef": surface.et0_coefficient,
                                    # 'crop_factor' : surface.crop_factor
                                }
                            )
                            for pol in constants.POLLUTANTS:
                                surfaces[-1][pol] = surface.storage[pol]
                        else:
                            surfaces.append(
                                {
                                    "node": name,
                                    "surface": surface.surface,
                                    "storage": surface.storage["volume"],
                                    "evaporation": surface.evaporation["volume"],
                                    "precipitation": surface.precipitation["volume"],
                                    "capacity": surface.capacity,
                                    "time": date,
                                }
                            )
                            for pol in constants.POLLUTANTS:
                                surfaces[-1][pol] = surface.storage[pol]

            for node in self.nodes.values():
                node.end_timestep()

            for arc in self.arcs.values():
                arc.end_timestep()
        objective_results = []
        for objective in objectives:
            if objective["element_type"] == "tanks":
                val = objective["function"](
                    [x for x in tanks if x["node"] == objective["name"]], self
                )
            elif objective["element_type"] == "flows":
                val = objective["function"](
                    [x for x in flows if x["arc"] == objective["name"]], self
                )
            elif objective["element_type"] == "surfaces":
                val = objective["function"](
                    [
                        x
                        for x in surfaces
                        if (x["node"] == objective["name"])
                        & (x["surface"] == objective["surface"])
                    ],
                    self,
                )
            objective_results.append(val)
        if not verbose:
            enablePrint(stdout)
        return flows, tanks, objective_results, surfaces

    def reinit(self):
        """Reinitialise by ending all node/arc timesteps and calling reinit function in
        all nodes (generally zero-ing their storage values)."""
        for node in self.nodes.values():
            node.end_timestep()
            for prop in dir(node):
                prop = node.__getattribute__(prop)
                for prop_ in dir(prop):
                    if prop_ == "reinit":
                        prop_ = node.__getattribute__(prop_)
                        prop_()

        for arc in self.arcs.values():
            arc.end_timestep()


def write_yaml(address, config_name, data):
    """

    Args:
        address:
        config_name:
        data:
    """
    with open(os.path.join(address, config_name), "w") as file:
        yaml.dump(
            data,
            file,
            default_flow_style=False,
            sort_keys=False,
            Dumper=yaml.SafeDumper,
        )


def open_func(file_path, mode):
    """

    Args:
        file_path:
        mode:

    Returns:

    """
    if mode == "rt" and file_path.endswith(".gz"):
        return gzip.open(file_path, mode)
    else:
        return open(file_path, mode)


def read_csv(file_path, delimiter=","):
    """

    Args:
        file_path:
        delimiter:

    Returns:

    """
    with open_func(file_path, "rt") as f:
        reader = csv.DictReader(f, delimiter=delimiter)
        data = {}
        for row in reader:
            key = (row["variable"], to_datetime(row["time"]))
            value = float(row["value"])
            data[key] = value
        return data


def write_csv(data, fixed_data={}, filename="", compress=False):
    """

    Args:
        data:
        fixed_data:
        filename:
        compress:
    """
    if compress:
        open_func = gzip.open
        mode = "wt"
    else:
        open_func = open
        mode = "w"
    with open_func(filename, mode, newline="") as csvfile:
        writer = csv.writer(csvfile)
        writer.writerow(list(fixed_data.keys()) + ["variable", "time", "value"])
        fixed_data_values = list(fixed_data.values())
        for key, value in data.items():
            writer.writerow(fixed_data_values + list(key) + [str(value)])


def flatten_dict(d, parent_key="", sep="-"):
    """

    Args:
        d:
        parent_key:
        sep:

    Returns:

    """
    # Initialize an empty dictionary
    flat_dict = {}
    # Loop through each key-value pair in the input dictionary
    for k, v in d.items():
        # Construct a new key by appending the parent key and separator
        new_key = str(parent_key) + sep + str(k) if parent_key else k
        # If the value is another dictionary, call the function recursively
        if isinstance(v, dict):
            flat_dict.update(flatten_dict(v, new_key, sep))
        # Otherwise, add the key-value pair to the flat dictionary
        else:
            flat_dict[new_key] = v
    # Return the flattened dictionary
    return flat_dict


def check_and_convert_string(value):
    """

    Args:
        value:

    Returns:

    """
    try:
        return int(value)
    except Exception:
        try:
            return float(value)
        except Exception:
            if value == "None":
                return None
            else:
                return value


def unflatten_dict(d, sep=":"):
    """

    Args:
        d:
        sep:

    Returns:

    """
    result = {}
    for k, v in d.items():
        keys = k.split(sep)
        current = result
        for key in keys[:-1]:
            current = current.setdefault(key, {})
        current[keys[-1]] = v
    return result


def convert_keys(d):
    """

    Args:
        d:

    Returns:

    """
    # base case: if d is not a dict, return d
    if not isinstance(d, dict):
        return d
    # recursive case: create a new dict with int keys and converted values
    new_d = {}
    for k, v in d.items():
        new_d[check_and_convert_string(k)] = convert_keys(v)
    return new_d


def csv2yaml(address, config_name="config_csv.yml", csv_folder_name="csv"):
    """

    Args:
        address:
        config_name:
        csv_folder_name:
    """
    csv_path = os.path.join(address, csv_folder_name)
    csv_list = [
        os.path.join(csv_path, f)
        for f in os.listdir(csv_path)
        if os.path.isfile(os.path.join(csv_path, f))
    ]
    objs_type = {"nodes": {}, "arcs": {}}
    for fid in csv_list:
        with open(fid, "rt") as f:
            if "Dates" in fid:
                reader = csv.reader(f, delimiter=",")
                dates = []
                for row in reader:
                    dates.append(row[0])
                objs_type["dates"] = dates[1:]
            else:
                reader = csv.DictReader(f, delimiter=",")
                data = {}
                for row in reader:
                    formatted_row = {}
                    for key, value in row.items():
                        if value:
                            if ("[" in value) & ("]" in value):
                                # Convert lists
                                value = value.strip("[]")  # Remove the brackets
                                value = value.replace("'", "")  # Remove the string bits
                                value = value.split(", ")  # Split by comma
                                value = [check_and_convert_string(x) for x in value]
                            else:
                                # Convert ints, floats and strings
                                value = check_and_convert_string(value)

                            # Convert key and store converted values
                            formatted_row[key] = value
                    if "Sim_params" not in fid:
                        label = formatted_row["label"]
                        del formatted_row["label"]

                    formatted_row = unflatten_dict(formatted_row)
                    formatted_row = convert_keys(formatted_row)

                    # Convert nested dicts dicts
                    data[row["name"]] = formatted_row
                if "Sim_params" in fid:
                    objs_type = {
                        **objs_type,
                        **{x: y["value"] for x, y in data.items()},
                    }
                else:
                    objs_type[label] = {**objs_type[label], **data}
    write_yaml(address, config_name, objs_type)


def yaml2csv(address, config_name="config.yml", csv_folder_name="csv"):
    """

    Args:
        address:
        config_name:
        csv_folder_name:
    """
    with open(os.path.join(address, config_name), "r") as file:
        data = yaml.safe_load(file)

    # Format to easy format to write to database
    objs_type = {}
    for objects, object_label in zip([data["nodes"], data["arcs"]], ["nodes", "arcs"]):
        for key, value in objects.items():
            if isinstance(value, dict):
                # Identify node type
                if "node_type_override" in value.keys():
                    type_ = value["node_type_override"]
                elif "type_" in value.keys():
                    type_ = value["type_"]
                else:
                    type_ = False

                if type_:
                    # Flatten dictionaries
                    new_dict = {}
                    if type_ not in objs_type.keys():
                        objs_type[type_] = {}

                    for key_, value_ in value.items():
                        if isinstance(value_, dict):
                            new_dict[key_] = flatten_dict(value_, key_, ":")

                    for key_, value_ in new_dict.items():
                        del value[key_]
                        value = {**value, **value_}
                    value["label"] = object_label
                    objs_type[type_][key] = value

    del data["nodes"]
    del data["arcs"]
    if "dates" in data.keys():
        objs_type["Dates"] = data["dates"]
        del data["dates"]

    objs_type["Sim_params"] = {x: {"name": x, "value": y} for x, y in data.items()}

    csv_dir = os.path.join(address, csv_folder_name)

    if not os.path.exists(csv_dir):
        os.mkdir(csv_dir)

    for key, value in objs_type.items():
        if key == "Sim_params":
            fields = ["name", "value"]
        elif key == "Dates":
            fields = ["date"]
        else:
            fields = {}
            for value_ in value.values():
                fields = {**fields, **value_}

            del fields["name"]
            fields = ["name"] + list(fields.keys())

        with open(
            os.path.join(csv_dir, "{0}.csv".format(key)), "w", newline=""
        ) as csvfile:
            writer = csv.writer(csvfile)
            writer.writerow(fields)
            if key == "Dates":
                for date in value:
                    writer.writerow([date])
            else:
                for key_, value_ in value.items():
                    writer.writerow(
                        [str(value_[x]) if x in value_.keys() else None for x in fields]
                    )


def load_extension_files(files: list[str]) -> None:
    """Load extension files from a list of files.

    Args:
        files (list[str]): List of file paths to load

    Raises:
        ValueError: If file is not a .py file
        FileNotFoundError: If file does not exist
    """
    import importlib
    from pathlib import Path

    for file in files:
        if not file.endswith(".py"):
            raise ValueError(f"Only .py files are supported. Invalid file: {file}")
        if not Path(file).exists():
            raise FileNotFoundError(f"File {file} does not exist")

        spec = importlib.util.spec_from_file_location("module.name", file)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
